import jax
import keras

from bayesflow.utils import filter_kwargs


class JAXApproximator(keras.Model):
    # noinspection PyMethodOverriding
    def compute_metrics(self, *args, **kwargs) -> dict[str, jax.Array]:
        # implemented by each respective architecture
        raise NotImplementedError

    def stateless_compute_metrics(
        self,
        trainable_variables: any,
        non_trainable_variables: any,
        metrics_variables: any,
        data: dict[str, any],
        stage: str = "training",
    ) -> (jax.Array, tuple):
        """
        Things we do for jax:
        1. Accept trainable variables as the first argument
            (can be at any position as indicated by the argnum parameter
             in autograd, but needs to be an explicit arg)
        2. Accept, potentially modify, and return other state variables
        3. Return just the loss tensor as the first value
        4. Return all other values in a tuple as the second value

        This ensures:
        1. The function is stateless
        2. The function can be differentiated with jax autograd
        """
        state_mapping = []
        state_mapping.extend(zip(self.trainable_variables, trainable_variables))
        state_mapping.extend(zip(self.non_trainable_variables, non_trainable_variables))
        state_mapping.extend(zip(self.metrics_variables, metrics_variables))

        # perform a stateless call to compute_metrics
        with keras.StatelessScope(state_mapping) as scope:
            kwargs = filter_kwargs(data | {"stage": stage}, self.compute_metrics)
            metrics = self.compute_metrics(**kwargs)

        # update variables
        non_trainable_variables = [scope.get_current_value(v) for v in self.non_trainable_variables]
        metrics_variables = [scope.get_current_value(v) for v in self.metrics_variables]

        return metrics["loss"], (metrics, non_trainable_variables, metrics_variables)

    def stateless_test_step(self, state: tuple, data: dict[str, any]) -> (dict[str, jax.Array], tuple):
        trainable_variables, non_trainable_variables, metrics_variables = state

        loss, aux = self.stateless_compute_metrics(
            trainable_variables, non_trainable_variables, metrics_variables, data=data, stage="validation"
        )
        metrics, non_trainable_variables, metrics_variables = aux

        metrics_variables = self._update_loss(loss, metrics_variables)

        state = trainable_variables, non_trainable_variables, metrics_variables
        return metrics, state

    def stateless_train_step(self, state: tuple, data: dict[str, any]) -> (dict[str, jax.Array], tuple):
        trainable_variables, non_trainable_variables, optimizer_variables, metrics_variables = state

        grad_fn = jax.value_and_grad(self.stateless_compute_metrics, has_aux=True)

        (loss, aux), grads = grad_fn(
            trainable_variables, non_trainable_variables, metrics_variables, data=data, stage="training"
        )
        metrics, non_trainable_variables, metrics_variables = aux

        trainable_variables, optimizer_variables = self.optimizer.stateless_apply(
            optimizer_variables, grads, trainable_variables
        )

        metrics_variables = self._update_loss(loss, metrics_variables)

        state = trainable_variables, non_trainable_variables, optimizer_variables, metrics_variables
        return metrics, state

    def test_step(self, *args, **kwargs):
        return self.stateless_test_step(*args, **kwargs)

    def train_step(self, *args, **kwargs):
        return self.stateless_train_step(*args, **kwargs)

    def _update_loss(self, loss: jax.Array, metrics_variables: any) -> any:
        # update the loss progress bar, and possibly metrics variables along with it
        state_mapping = list(zip(self.metrics_variables, metrics_variables))
        with keras.StatelessScope(state_mapping) as scope:
            self._loss_tracker.update_state(loss)

        metrics_variables = [scope.get_current_value(v) for v in self.metrics_variables]

        return metrics_variables



from collections.abc import Sequence, Mapping

import matplotlib.pyplot as plt

import numpy as np
import pandas as pd
import seaborn as sns

from bayesflow.utils.dict_utils import dicts_to_arrays

from .pairs_samples import _pairs_samples


def pairs_posterior(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray = None,
    priors: Mapping[str, np.ndarray] | np.ndarray = None,
    dataset_id: int = None,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    height: int = 3,
    post_color: str | tuple = "#132a70",
    prior_color: str | tuple = "gray",
    alpha: float = 0.9,
    label_fontsize: int = 14,
    tick_fontsize: int = 12,
    legend_fontsize: int = 14,
    **kwargs,
) -> sns.PairGrid:
    """Generates a bivariate pair plot given posterior draws and optional prior or prior draws.

    Parameters
    ----------
    estimates   : np.ndarray of shape (n_post_draws, n_params)
        The posterior draws obtained for a SINGLE observed data set.
    targets       : np.ndarray of shape (n_params,) or None, optional, default: None
        Optional true parameter values that have generated the observed dataset.
    priors       : np.ndarray of shape (n_prior_draws, n_params) or None, optional (default: None)
        Optional prior samples obtained from the prior.
    dataset_id: Optional ID of the dataset for whose posterior the pairs plot shall be generated.
        Should only be specified if estimates contains posterior draws from multiple datasets.
    variable_keys       : list or None, optional, default: None
       Select keys from the dictionary provided in samples.
       By default, select all keys.
    variable_names       : list or None, optional, default: None
        The parameter names for nice plot titles. Inferred if None
    height            : float, optional, default: 3
        The height of the pairplot
    label_fontsize    : int, optional, default: 14
        The font size of the x and y-label texts (parameter names)
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels
    legend_fontsize   : int, optional, default: 16
        The font size of the legend text
    post_color        : str, optional, default: '#132a70'
        The color for the posterior histograms and KDEs
    prior_color      : str, optional, default: gray
        The color for the optional prior histograms and KDEs
    alpha             : float in [0, 1], optional, default: 0.9
        The opacity of the posterior plots

    **kwargs          : dict, optional, default: {}
        Further optional keyword arguments propagated to `_pairs_samples`

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    AssertionError
        If the shape of posterior_draws is not 2-dimensional.
    """

    plot_data = dicts_to_arrays(
        estimates=estimates,
        targets=targets,
        priors=priors,
        dataset_ids=dataset_id,
        variable_keys=variable_keys,
        variable_names=variable_names,
    )

    # dicts_to_arrays will keep dataset axis even if it is of length 1
    # however, pairs plotting requires the dataset axis to be removed
    estimates_shape = plot_data["estimates"].shape
    if len(estimates_shape) == 3 and estimates_shape[0] == 1:
        plot_data["estimates"] = np.squeeze(plot_data["estimates"], axis=0)

    # plot posterior first
    g = _pairs_samples(
        plot_data=plot_data,
        height=height,
        color=post_color,
        color2=prior_color,
        alpha=alpha,
        label_fontsize=label_fontsize,
        tick_fontsize=tick_fontsize,
        legend_fontsize=legend_fontsize,
        **kwargs,
    )

    targets = plot_data.get("targets")
    if targets is not None:
        # Ensure targets is at least 2D
        if targets.ndim == 1:
            targets = np.atleast_2d(targets)

        # Create DataFrame with variable names as columns
        g.data = pd.DataFrame(targets, columns=targets.variable_names)
        g.data["_source"] = "True Parameter"
        g.map_diag(plot_true_params)

    return g


def plot_true_params(x, hue=None, **kwargs):
    """Custom function to plot true parameters on the diagonal."""
    # hue needs to be added to handle the case of plotting both posterior and prior
    param = x.iloc[0]  # Get the single true value for the diagonal
    # only plot on the diagonal a vertical line for the true parameter
    plt.axvline(param, color="black", linestyle="--")



import keras

from bayesflow.types import Tensor
from bayesflow.utils import expand_tile
from bayesflow.utils.serialization import serializable


@serializable
class Time2Vec(keras.Layer):
    """
    Implements the Time2Vec learnbale embedding from [1].
    [1] Kazemi, S. M., Goel, R., Eghbali, S., Ramanan, J., Sahota, J., Thakur, S., ... & Brubaker, M.
    (2019). Time2vec: Learning a vector representation of time. arXiv preprint arXiv:1907.05321.
    """

    def __init__(self, num_periodic_features: int = 8):
        """
        Initializes a time series decomposition model with learnable trend and periodic components.

        This model represents a time-dependent signal as a combination of a linear trend and periodic features.

        The trend is parameterized by a learnable weight and bias, while the  periodic component consists of multiple
        sine terms with trainable weights and biases. The number of periodic features determines the complexity of the
        periodic representation.

        Parameters
        ----------
        num_periodic_features : int, optional
            The number of periodic components used in the decomposition. Higher values allow
            capturing more complex periodic patterns. Default is 8.

        """

        super().__init__()

        self.num_periodic_features = num_periodic_features
        self.linear_weight = self.add_weight(
            shape=(1,),
            initializer="glorot_uniform",
            trainable=True,
            name="trend_weight",
        )

        self.linear_bias = self.add_weight(
            shape=(1,),
            initializer="glorot_uniform",
            trainable=True,
            name="trend_bias",
        )

        self.periodic_weights = self.add_weight(
            shape=(self.num_periodic_features,),
            initializer="glorot_normal",
            trainable=True,
            name="periodic_weights",
        )

        self.periodic_biases = self.add_weight(
            shape=(self.num_periodic_features,),
            initializer="glorot_normal",
            trainable=True,
            name="periodic_biases",
        )

    def call(self, x: Tensor, t: Tensor = None) -> Tensor:
        """Creates time representations and concatenates them to x.

        Parameters
        ----------
        x   : Tensor of shape (batch_size, sequence_length, dim)
            The input sequence.
        t   : Tensor of shape (batch_size, sequence_length)
            Vector of times

        Returns
        -------
        emb : Tensor
            Embedding of shape (batch_size, sequence_length, num_periodic_features + 1)
        """

        if t is None:
            t = keras.ops.linspace(0, keras.ops.shape(x)[1], keras.ops.shape(x)[1], dtype=x.dtype)
            t = expand_tile(t, keras.ops.shape(x)[0], axis=0)

        linear = t * self.linear_weight + self.linear_bias
        periodic = keras.ops.sin(t[..., None] * self.periodic_weights[None, :] + self.periodic_biases[None, :])
        emb = keras.ops.concatenate([linear[..., None], periodic], axis=-1)
        return keras.ops.concatenate([x, emb], axis=-1)



import keras
import torch

from bayesflow.utils import filter_kwargs


class TorchApproximator(keras.Model):
    # noinspection PyMethodOverriding
    def compute_metrics(self, *args, **kwargs) -> dict[str, torch.Tensor]:
        # implemented by each respective architecture
        raise NotImplementedError

    def test_step(self, data: dict[str, any]) -> dict[str, torch.Tensor]:
        kwargs = filter_kwargs(data | {"stage": "validation"}, self.compute_metrics)
        return self.compute_metrics(**kwargs)

    def train_step(self, data: dict[str, any]) -> dict[str, torch.Tensor]:
        with torch.enable_grad():
            kwargs = filter_kwargs(data | {"stage": "training"}, self.compute_metrics)
            metrics = self.compute_metrics(**kwargs)

        loss = metrics["loss"]

        # noinspection PyUnresolvedReferences
        self.zero_grad()
        loss.backward()

        trainable_weights = self.trainable_weights[:]
        gradients = [v.value.grad for v in trainable_weights]

        # Update weights
        with torch.no_grad():
            self.optimizer.apply(gradients, trainable_weights)

        self._loss_tracker.update_state(loss)

        return metrics



import numpy as np

from .benchmark_simulator import BenchmarkSimulator


class InverseKinematics(BenchmarkSimulator):
    def __init__(
        self,
        scales: np.ndarray = None,
        l1: float = 0.5,
        l2: float = 0.5,
        l3: float = 1.0,
        rng: np.random.Generator = None,
    ):
        """Inverse Kinematics simulated benchmark
        See: https://arxiv.org/pdf/2101.10763.pdf

        Parameters
        ----------
        scales: np.ndarray of shape (4, ) or None, optional, default: None
            The four scales of the Gaussian prior.
            If ``None`` provided, the scales from https://arxiv.org/pdf/2101.10763.pdf
            will be used: [0.25, 0.5, 0.5, 0.5].
        l1: float, optional, default: 0.5
            The length of the first segment.
        l2: float, optional, default: 0.5
            The length of the second segment.
        l3: float, optional, default: 1.0
            The length of the third segment.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.scales = scales
        if self.scales is None:
            self.scales = np.array([0.25, 0.5, 0.5, 0.5])
        self.l1 = l1
        self.l2 = l2
        self.l3 = l3
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

    def prior(self):
        """Generates a random draw from a 4-dimensional Gaussian prior distribution with a
        spherical convariance matrix. The parameters represent a robot's arm configuration,
        with the first parameter indicating the arm's height and the remaining three are
        angles.

        Returns
        -------
        params : np.ndarray of shape (4, )
            A single draw from the 4-dimensional Gaussian prior.
        """

        return self.rng.normal(loc=0, scale=self.scales)

    def observation_model(self, params: np.ndarray):
        """Returns the 2D coordinates of a robot arm given parameter vector.
        The first parameter represents the arm's height and the remaining three
        correspond to angles.

        Parameters
        ----------
        params   : np.ndarray of shape (params, )
            The four model parameters which will determine the coordinates

        Returns
        -------
        x : np.ndarray of shape (2, )
            The 2D coordinates of the arm
        """

        # Determine 2D position
        x1 = self.l1 * np.sin(params[1])
        x1 += self.l2 * np.sin(params[1] + params[2])
        x1 += self.l3 * np.sin(params[1] + params[2] + params[3]) + params[0]
        x2 = self.l1 * np.cos(params[1])
        x2 += self.l2 * np.cos(params[1] + params[2])
        x2 += self.l3 * np.cos(params[1] + params[2] + params[3])
        return np.array([x1, x2])



import keras
from keras.saving import (
    deserialize_keras_object as deserialize,
    serialize_keras_object as serialize,
    register_keras_serializable as serializable,
)

from bayesflow.utils import model_kwargs, find_network, serialize_value_or_type, deserialize_value_or_type
from bayesflow.types import Shape, Tensor
from bayesflow.scores import ScoringRule, ParametricDistributionScore
from bayesflow.utils.decorators import allow_batch_size


@serializable(package="networks.point_inference_network")
class PointInferenceNetwork(keras.Layer):
    """Implements point estimation for user specified scoring rules by a shared feed forward architecture
    with separate heads for each scoring rule.
    """

    def __init__(
        self,
        scores: dict[str, ScoringRule],
        subnet: str | keras.Layer = "mlp",
        **kwargs,
    ):
        super().__init__(**model_kwargs(kwargs))

        self.scores = scores

        self.subnet = find_network(subnet, **kwargs.get("subnet_kwargs", {}))

        self.config = {
            **kwargs,
        }
        self.config = serialize_value_or_type(self.config, "subnet", subnet)
        self.config["scores"] = serialize(self.scores)

    def build(self, xz_shape: Shape, conditions_shape: Shape = None) -> None:
        """Builds all network components based on shapes of conditions and targets.

        For each score, corresponding estimation heads are constructed.
        There are two steps in this:

        #. Request a dictionary of names and output shapes of required heads from the score.
        #. Then for each required head, request corresponding head networks from the score.

        Since the score is in charge of constructing heads, this allows for convenient yet flexible building.
        """
        if conditions_shape is None:  # unconditional estimation uses a fixed input vector
            input_shape = (1, 1)
        else:
            input_shape = conditions_shape

        # Save input_shape and xz_shape for usage in get_build_config
        self._input_shape = input_shape
        self._xz_shape = xz_shape

        # build the shared body network
        self.subnet.build(input_shape)
        body_output_shape = self.subnet.compute_output_shape(input_shape)

        # build head(s) for every scoring rule
        self.heads = dict()
        self.heads_flat = dict()  # see comment regarding heads_flat below

        for score_key, score in self.scores.items():
            head_shapes = score.get_head_shapes_from_target_shape(xz_shape)

            self.heads[score_key] = {}

            for head_key, head_shape in head_shapes.items():
                head = score.get_head(head_key, head_shape)
                head.build(body_output_shape)
                # If head is not tracked explicitly, self.variables does not include them.
                # Testing with tests.utils.assert_layers_equal() would thus neglect heads
                head = self._tracker.track(head)  # explicitly track head

                self.heads[score_key][head_key] = head

                # Until keras issue [20598](https://github.com/keras-team/keras/issues/20598)
                # is resolved, a flat version of the heads dictionary is kept.
                # This allows to save head weights properly, see for reference
                # https://github.com/keras-team/keras/blob/v3.3.3/keras/src/saving/saving_lib.py#L481.
                # A nested heads dict is still preferred over this flat dict,
                # because it avoids string operation based filtering in `self._forward()`.
                flat_key = f"{score_key}___{head_key}"
                self.heads_flat[flat_key] = head

    def get_build_config(self):
        build_config = {
            "conditions_shape": self._input_shape,
            "xz_shape": self._xz_shape,
        }

        # Save names of head networks
        heads = {}
        for score_key in self.heads.keys():
            heads[score_key] = {}
            for head_key, head in self.heads[score_key].items():
                heads[score_key][head_key] = head.name

        build_config["heads"] = heads

        return build_config

    def build_from_config(self, config):
        self.build(xz_shape=config["xz_shape"], conditions_shape=config["conditions_shape"])

        for score_key in self.scores.keys():
            for head_key, head in self.heads[score_key].items():
                head.name = config["heads"][score_key][head_key]

    def get_config(self):
        base_config = super().get_config()

        return base_config | self.config

    @classmethod
    def from_config(cls, config):
        config = config.copy()
        config["scores"] = deserialize(config["scores"])
        config = deserialize_value_or_type(config, "subnet")
        return cls(**config)

    def call(
        self,
        xz: Tensor = None,
        conditions: Tensor = None,
        training: bool = False,
        **kwargs,
    ) -> dict[str, dict[str, Tensor]]:
        if xz is None and not self.built:
            raise ValueError("Cannot build inference network without inference variables.")
        if conditions is None:  # unconditional estimation uses a fixed input vector
            conditions = keras.ops.convert_to_tensor(
                [[1.0]], dtype=keras.ops.dtype(xz) if xz is not None else "float32"
            )

        # pass conditions to the shared subnet
        output = self.subnet(conditions, training=training)

        # pass along to calculate individual head outputs
        output = {
            score_key: {head_key: head(output, training=training) for head_key, head in self.heads[score_key].items()}
            for score_key in self.heads.keys()
        }
        return output

    def compute_metrics(
        self, x: Tensor, conditions: Tensor = None, sample_weight: Tensor = None, stage: str = "training"
    ) -> dict[str, Tensor]:
        output = self(x, conditions)

        metrics = {}
        # calculate negative score as mean over all scores
        for score_key, score in self.scores.items():
            score_value = score.score(output[score_key], x, sample_weight)
            metrics[score_key] = score_value
        neg_score = keras.ops.mean(list(metrics.values()))

        if stage != "training" and any(self.metrics):
            # compute sample-based metrics
            samples = self.sample((keras.ops.shape(x)[0],), conditions=conditions)

            for metric in self.metrics:
                metrics[metric.name] = metric(samples, x)

        return metrics | {"loss": neg_score}

    @allow_batch_size
    def sample(self, batch_shape: Shape, conditions: Tensor = None) -> dict[str, Tensor]:
        """
        Parameters
        ----------
        batch_shape : tuple,
            Expected dimensions depend on `conditions`
            - conditional sampling: (batch_size, num_samples) if `conditions` is a tensor
              of shape (batch_size, num_samples)
            - unconditional sampling: (num_samples,) if `conditions` is None
        conditions : Tensor or None, default None
            Optional inference conditions. If `conditions` is not given, the method will return unconditional samples.

        Returns
        -------
        samples : dict[str, Tensor]
            Samples for every parametric scoring rule. Dict values have shape (batch_size, num_samples, num_variables)
            or (num_samples, num_variables) for conditional or unconditional sampling respectively.
        """
        if conditions is None:  # unconditional estimation uses a fixed input vector
            conditions = keras.ops.ones(batch_shape, dtype="float32").reshape(1, -1, 1)

        # conditions are duplicated along axis 1 num_sample times
        output = self.subnet(conditions[:, 0, :])
        samples = {}

        for score_key, score in self.scores.items():
            if isinstance(score, ParametricDistributionScore):
                parameters = {head_key: head(output) for head_key, head in self.heads[score_key].items()}
                samples[score_key] = score.sample(batch_shape, **parameters)

        return samples

    def log_prob(self, samples: Tensor, conditions: Tensor = None, **kwargs) -> dict[str, Tensor]:
        output = self.subnet(conditions)
        log_probs = {}

        for score_key, score in self.scores.items():
            if isinstance(score, ParametricDistributionScore):
                parameters = {head_key: head(output) for head_key, head in self.heads[score_key].items()}
                log_probs[score_key] = score.log_prob(x=samples, **parameters)

        return log_probs



from collections.abc import Callable
import keras

from bayesflow.utils import model_kwargs
from bayesflow.utils.serialization import deserialize, serializable, serialize


@serializable
class DoubleConv(keras.Sequential):
    def __init__(
        self,
        width: int,
        use_batchnorm: bool = True,
        dropout: float = None,
        activation: str | Callable[[], keras.Layer] = "relu",
        **kwargs,
    ):
        layers = []

        layers.append(keras.layers.Conv2D(width, 3, padding="same"))

        activation = keras.activations.get(activation)
        if not isinstance(activation, keras.Layer):
            activation = keras.layers.Activation(activation)

        layers.append(activation)

        if use_batchnorm:
            # we apply this after the activation due to
            # https://github.com/keras-team/keras/issues/1802#issuecomment-187966878
            layers.append(keras.layers.BatchNormalization())

        if dropout is not None and dropout > 0:
            layers.append(keras.layers.Dropout(dropout))

        layers.append(keras.layers.Conv2D(width, 3, padding="same"))

        if use_batchnorm:
            layers.append(keras.layers.BatchNormalization())

        super().__init__(layers, **model_kwargs(kwargs))

        self.width = width
        self.use_batchnorm = use_batchnorm
        self.dropout = dropout
        self.activation = activation

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()

        config = {
            "width": self.width,
            "use_batchnorm": self.use_batchnorm,
            "dropout": self.dropout,
            "activation": self.activation,
        }

        return base_config | serialize(config)

    def build(self, input_shape=None):
        # set the padding so the output is max-poolable
        *batch_shape, height, width, channels = input_shape

        padding = [height % 2, width % 2]
        self._layers.insert(0, keras.layers.ZeroPadding2D(padding=padding))

        super().build(input_shape)



from .find_cost import find_cost
from .find_network import find_network
from .find_permutation import find_permutation
from .find_pooling import find_pooling
from .find_recurrent_net import find_recurrent_net
from .find_inference_network import find_inference_network
from .find_summary_network import find_summary_network
from .find_distribution import find_distribution



from collections.abc import Callable, Sequence
import keras

from bayesflow.utils import model_kwargs
from bayesflow.utils.serialization import deserialize, serializable, serialize

from bayesflow.networks.residual import Residual
from .double_linear import DoubleLinear


@serializable
class DenseResNet(keras.Sequential):
    """
    Implements the fully-connected analogue of the ResNet architecture.

    Note that we still apply dropout and activation to the output,
    so you will need to apply at least one linear layer after this network.
    """

    def __init__(
        self,
        widths: Sequence[int],
        use_batchnorm: bool = True,
        dropout: float = 0.05,
        activation: str | Callable[[], keras.Layer] = "relu",
        **kwargs,
    ):
        layers = []

        for width in widths:
            layer = DoubleLinear(width, use_batchnorm=use_batchnorm, dropout=dropout, activation=activation)
            layer = Residual(layer)
            activation = keras.activations.get(activation)
            if not isinstance(activation, keras.Layer):
                activation = keras.layers.Activation(activation)

            layers.append(layer)
            layers.append(activation)

        super().__init__(layers, **model_kwargs(kwargs))

        self.widths = widths
        self.use_batchnorm = use_batchnorm
        self.dropout = dropout
        self.activation = activation

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        base_config = super().get_config()

        config = {
            "widths": self.widths,
            "use_batchnorm": self.use_batchnorm,
            "dropout": self.dropout,
            "activation": self.activation,
        }

        return base_config | serialize(config)



from collections.abc import Sequence
import numpy as np

from bayesflow.utils.serialization import deserialize, serializable, serialize

from .transform import Transform


@serializable
class Split(Transform):
    """This is the effective inverse of the :py:class:`~Concatenate` Transform.

    Parameters
    ----------
    key : str
        The key to split in the forward transform.
    into: Sequence[str]
        The names of each split after the forward transform.
    indices_or_sections : int | Sequence[int], optional, default: None
        The number of sections or indices to split on. If not given, will split evenly into len(into) parts.
    axis: int, optional, default: -1
        The axis to split on.
    """

    def __init__(self, key: str, into: Sequence[str], indices_or_sections: int | Sequence[int] = None, axis: int = -1):
        self.axis = axis
        self.key = key
        self.into = into

        if indices_or_sections is None:
            indices_or_sections = len(into)

        self.indices_or_sections = indices_or_sections

    @classmethod
    def from_config(cls, config: dict, custom_objects=None) -> "Split":
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self) -> dict:
        config = {
            "key": self.key,
            "into": self.into,
            "indices_or_sections": self.indices_or_sections,
            "axis": self.axis,
        }
        return serialize(config)

    def forward(self, data: dict[str, np.ndarray], strict: bool = True, **kwargs) -> dict[str, np.ndarray]:
        # avoid side effects
        data = data.copy()

        if strict and self.key not in data:
            raise KeyError(self.key)
        elif self.key not in data:
            # we cannot produce a result, but also don't have to
            return data

        splits = np.split(data.pop(self.key), self.indices_or_sections, axis=self.axis)

        if len(splits) != len(self.into):
            raise ValueError(f"Requested {len(self.into)} splits, but produced {len(splits)}.")

        for key, split in zip(self.into, splits):
            data[key] = split

        return data

    def inverse(self, data: dict[str, np.ndarray], strict: bool = False, **kwargs) -> dict[str, np.ndarray]:
        # avoid side effects
        data = data.copy()

        required_keys = set(self.into)
        available_keys = set(data.keys())
        common_keys = available_keys & required_keys
        missing_keys = required_keys - available_keys

        if strict and missing_keys:
            # invalid call
            raise KeyError(f"Missing keys: {missing_keys!r}")
        elif missing_keys:
            # we cannot produce a result, but should still remove the keys
            for key in common_keys:
                data.pop(key)

            return data

        # remove each part
        splits = [data.pop(key) for key in self.into]

        # concatenate them all
        result = np.concatenate(splits, axis=self.axis)

        # store the result
        data[self.key] = result

        return data

    def extra_repr(self) -> str:
        result = "[" + ", ".join(map(repr, self.key)) + "] -> " + repr(self.into)

        if self.axis != -1:
            result += f", axis={self.axis}"

        return result



"""
A rich collection of neural network architectures for use in :py:class:`~bayesflow.approximators.Approximator`\ s.
"""

from .consistency_models import ConsistencyModel
from .coupling_flow import CouplingFlow
from .deep_set import DeepSet
from .flow_matching import FlowMatching
from .inference_network import InferenceNetwork
from .point_inference_network import PointInferenceNetwork
from .mlp import MLP
from .summary_network import SummaryNetwork
from .time_series_network import TimeSeriesNetwork
from .transformers import SetTransformer, TimeSeriesTransformer, FusionTransformer

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import pytest


@pytest.fixture()
def actnorm():
    from bayesflow.networks.coupling_flow.actnorm import ActNorm

    return ActNorm()


@pytest.fixture()
def dual_coupling(request, transform):
    from bayesflow.networks.coupling_flow.couplings import DualCoupling

    return DualCoupling(transform=transform)


@pytest.fixture(params=["actnorm", "dual_coupling"])
def invertible_layer(request, transform):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def single_coupling(request, transform):
    from bayesflow.networks.coupling_flow.couplings import SingleCoupling

    return SingleCoupling(transform=transform)


@pytest.fixture(params=["affine", "spline"])
def transform(request):
    return request.param



import math

import keras
from keras.saving import register_keras_serializable as serializable

from bayesflow.types import Shape, Tensor
from bayesflow.links import PositiveDefinite

from .parametric_distribution_score import ParametricDistributionScore


@serializable(package="bayesflow.scores")
class MultivariateNormalScore(ParametricDistributionScore):
    r""":math:`S(\hat p_{\mu, \Sigma}, \theta; k) = -\log( \mathcal N (\theta; \mu, \Sigma))`

    Scores a predicted mean and covariance matrix with the log-score of the probability of the materialized value.
    """

    NOT_TRANSFORMING_LIKE_VECTOR_WARNING = ("covariance",)
    """
    Marks head for covariance matrix as an exception for adapter transformations.

    This variable contains names of prediction heads that should lead to a warning when the adapter is applied
    in inverse direction to them.

    For more information see :py:class:`ScoringRule`.
    """

    def __init__(self, dim: int = None, links: dict = None, **kwargs):
        super().__init__(links=links, **kwargs)

        self.dim = dim
        self.links = links or {"covariance": PositiveDefinite()}

        self.config = {"dim": dim}

    def get_config(self):
        base_config = super().get_config()
        return base_config | self.config

    def get_head_shapes_from_target_shape(self, target_shape: Shape) -> dict[str, Shape]:
        self.dim = target_shape[-1]
        return dict(mean=(self.dim,), covariance=(self.dim, self.dim))

    def log_prob(self, x: Tensor, mean: Tensor, covariance: Tensor) -> Tensor:
        """
        Compute the log probability density of a multivariate Gaussian distribution.

        This function calculates the log probability density for each sample in `x` under a
        multivariate Gaussian distribution with the given `mean` and `covariance`.

        The computation includes the determinant of the covariance matrix, its inverse, and the quadratic
        form in the exponential term of the Gaussian density function.

        Parameters
        ----------
        x : Tensor
            A tensor of input samples for which the log probability density is computed.
            The shape should be compatible with broadcasting against `mean`.
        mean : Tensor
            A tensor representing the mean of the multivariate Gaussian distribution.
        covariance : Tensor
            A tensor representing the covariance matrix of the multivariate Gaussian distribution.

        Returns
        -------
        Tensor
            A tensor containing the log probability densities for each sample in `x` under the
            given Gaussian distribution.
        """
        diff = x - mean
        precision = keras.ops.inv(covariance)
        log_det_covariance = keras.ops.slogdet(covariance)[1]  # Only take the log of the determinant part

        # Compute the quadratic term in the exponential of the multivariate Gaussian
        quadratic_term = keras.ops.einsum("...i,...ij,...j->...", diff, precision, diff)

        # Compute the log probability density
        log_prob = -0.5 * (self.dim * keras.ops.log(2 * math.pi) + log_det_covariance + quadratic_term)

        return log_prob

    def sample(self, batch_shape: Shape, mean: Tensor, covariance: Tensor) -> Tensor:
        """
        Generate samples from a multivariate Gaussian distribution.

        This function samples from a multivariate Gaussian distribution with the given `mean`
        and `covariance` using the Cholesky decomposition method. Independent standard normal
        samples are transformed using the Cholesky factor of the covariance matrix to generate
        correlated samples.

        Parameters
        ----------
        batch_shape : Shape
            A tuple specifying the batch size and the number of samples to generate.
        mean : Tensor
            A tensor representing the mean of the multivariate Gaussian distribution.
            Must have shape (batch_size, D), where D is the dimensionality of the distribution.
        covariance : Tensor
            A tensor representing the covariance matrix of the multivariate Gaussian distribution.
            Must have shape (batch_size, D, D), where D is the dimensionality.

        Returns
        -------
        Tensor
            A tensor of shape (batch_size, num_samples, D) containing the generated samples.
        """
        if len(batch_shape) == 1:
            batch_shape = (1,) + tuple(batch_shape)
        batch_size, num_samples = batch_shape
        dim = keras.ops.shape(mean)[-1]
        if keras.ops.shape(mean) != (batch_size, dim):
            raise ValueError(f"mean must have shape (batch_size, {dim}), but got {keras.ops.shape(mean)}")

        if keras.ops.shape(covariance) != (batch_size, dim, dim):
            raise ValueError(
                f"covariance must have shape (batch_size, {dim}, {dim}), but got {keras.ops.shape(covariance)}"
            )

        # Use Cholesky decomposition to generate samples
        cholesky_factor = keras.ops.cholesky(covariance)
        normal_samples = keras.random.normal((*batch_shape, dim))

        scaled_normal = keras.ops.einsum("ijk,ilk->ilj", cholesky_factor, normal_samples)
        samples = mean[:, None, :] + scaled_normal

        return samples



from collections.abc import Sequence

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

import keras.src.callbacks

from ...utils.plot_utils import make_figure, add_titles_and_labels


def loss(
    history: keras.callbacks.History,
    train_key: str = "loss",
    val_key: str = "val_loss",
    per_training_step: bool = False,
    smoothing_factor: float = 0.8,
    figsize: Sequence[float] = None,
    train_color: str = "#132a70",
    val_color: str = "black",
    val_marker: str = "o",
    val_marker_size: float = 5,
    lw_train: float = 2.0,
    lw_val: float = 2.0,
    grid_alpha: float = 0.2,
    legend_fontsize: int = 14,
    label_fontsize: int = 14,
    title_fontsize: int = 16,
) -> plt.Figure:
    """
    A generic helper function to plot the losses of a series of training epochs and runs.

    Parameters
    ----------

    history     : keras.src.callbacks.History
        History object as returned by `keras.Model.fit`.
    train_key   : str, optional, default: "loss"
        The training loss key to look for in the history
    val_key     : str, optional, default: "val_loss"
        The validation loss key to look for in the history
    per_training_step : bool, optional, default: False
        A flag for making loss trajectory detailed (to training steps) rather than per epoch.
    smoothing_factor : float, optional, default: 0.8
        If greater than zero, smooth the loss curves by applying an exponential moving average.
    figsize            : tuple or None, optional, default: None
        The figure size passed to the ``matplotlib`` constructor.
        Inferred if ``None``
    train_color        : str, optional, default: '#132a70'
        The color for the train loss trajectory
    val_color          : str, optional, default: None
        The color for the optional validation loss trajectory
    val_marker: str
        Marker style for the validation loss curve. Default is "o".
    val_marker_size: float
        Marker size for the validation loss curve. Default is 5.
    lw_train           : int, optional, default: 2
        The line width for the training loss curve
    lw_val             : int, optional, default: 2
        The line width for the validation loss curve
    grid_alpha          : float, optional, default: 0.2
        The transparency of the background grid
    legend_fontsize    : int, optional, default: 14
        The font size of the legend text
    label_fontsize     : int, optional, default: 14
        The font size of the y-label text
    title_fontsize     : int, optional, default: 16
        The font size of the title text

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    AssertionError
        If the number of columns in ``train_losses`` does not match the
        number of columns in ``val_losses``.
    """

    train_losses = history.history.get(train_key)
    val_losses = history.history.get(val_key)

    train_losses = pd.DataFrame(np.array(train_losses))
    val_losses = pd.DataFrame(np.array(val_losses)) if val_losses is not None else None

    # Determine the number of rows for plot
    num_row = len(train_losses.columns)

    # Initialize figure
    fig, axes = make_figure(num_row=num_row, num_col=1, figsize=(16, int(4 * num_row)) if figsize is None else figsize)

    # Get the number of steps as an array
    train_step_index = np.arange(1, len(train_losses) + 1)
    if val_losses is not None:
        val_step = int(np.floor(len(train_losses) / len(val_losses)))
        val_step_index = train_step_index[(val_step - 1) :: val_step]

        # If unequal length due to some reason, attempt a fix
        if val_step_index.shape[0] > val_losses.shape[0]:
            val_step_index = val_step_index[: val_losses.shape[0]]

    # Loop through loss entries and populate plot
    for i, ax in enumerate(axes.flat):
        if smoothing_factor > 0:
            # plot unsmoothed train loss
            ax.plot(
                train_step_index, train_losses.iloc[:, 0], color=train_color, lw=lw_train, alpha=0.3, label="Training"
            )

            # plot smoothed train loss
            smoothed_train_loss = train_losses.iloc[:, 0].ewm(alpha=1.0 - smoothing_factor, adjust=True).mean()
            ax.plot(
                train_step_index,
                smoothed_train_loss,
                color=train_color,
                lw=lw_train,
                alpha=0.8,
                label="Training (Moving Average)",
            )
        else:
            # Plot unsmoothed train loss
            ax.plot(
                train_step_index, train_losses.iloc[:, 0], color=train_color, lw=lw_train, alpha=0.8, label="Training"
            )

        # Only plot if we actually have validation losses and a color assigned
        if val_losses is not None and val_color is not None:
            alpha_unsmoothed = 0.3 if smoothing_factor > 0 else 0.8

            # Plot unsmoothed val loss
            ax.plot(
                val_step_index,
                val_losses.iloc[:, 0],
                color=val_color,
                lw=lw_val,
                alpha=alpha_unsmoothed,
                linestyle="--",
                marker=val_marker,
                markersize=val_marker_size,
                label="Validation",
            )

            # if requested, plot a second, smoothed curve
            if smoothing_factor > 0:
                smoothed_val_loss = val_losses.iloc[:, 0].ewm(alpha=1.0 - smoothing_factor, adjust=True).mean()
                ax.plot(
                    val_step_index,
                    smoothed_val_loss,
                    color=val_color,
                    linestyle="--",
                    lw=lw_val,
                    alpha=0.8,
                    label="Validation (Moving Average)",
                )

        # rest of the styling
        sns.despine(ax=ax)
        ax.grid(alpha=grid_alpha)
        ax.set_xlim(train_step_index[0], train_step_index[-1])

        # legend only if there's at least one validation curve or smoothing was on
        if val_losses is not None or smoothing_factor > 0:
            ax.legend(fontsize=legend_fontsize)

    # Add labels, titles, and set font sizes
    add_titles_and_labels(
        axes=axes,
        num_row=num_row,
        num_col=1,
        title=["Loss Trajectory"],
        xlabel="Training step #" if per_training_step else "Training epoch #",
        ylabel="Value",
        title_fontsize=title_fontsize,
        label_fontsize=label_fontsize,
    )

    fig.tight_layout()
    return fig



from collections.abc import Sequence
import numpy as np
from scipy.stats import binom as scipy_binomial

from .minimal_coverage_probs import minimal_coverage_probs


def simultaneous_ecdf_bands(
    num_estimates: int,
    num_points: int = None,
    num_simulations: int = 1000,
    confidence: float = 0.95,
    eps: float = 1e-5,
    max_num_points: int = 1000,
) -> Sequence:
    """Computes the simultaneous ECDF bands through simulation according to
    the algorithm described in Section 2.2 of

    Säilynoja, T., Bürkner, P. C., & Vehtari, A. (2022).
    Graphical test for discrete uniformity and its applications in goodness-of-fit
    evaluation and multiple sample comparison. Statistics and Computing, 32(2), 32.
    See: https://link.springer.com/article/10.1007/s11222-022-10090-6

    Depends on the vectorized utility function `ecdf.minimal_coverage_probs(z, u)`.
    Will be used by the diagnostics module to create the ECDF marginal calibration plots.

    Parameters
    ----------
    num_estimates   : int
        The sample size used for computing the ECDF. Will equal to the number of simulated conditions when used
        for simulation-based calibration. Corresponds to `N` in the paper above.
    num_points      : int, optional, default: None
        The number of evaluation points on the interval (0, 1). Defaults to `num_points = num_estimates` if
        not explicitly specified. Correspond to `K` in the paper above.
    num_simulations : int, optional, default: 1000
        The number of samples of size `n_samples` to simulate for determining the simultaneous CIs.
    confidence      : float in (0, 1), optional, default: 0.95
        The confidence level, `confidence = 1 - alpha` specifies the width of the confidence interval.
    eps             : float, optional, default: 1e-5
        Small number to add to the lower and subtract from the upper bound of the interval [0, 1]
        to avoid edge artefacts. No need to touch this.
    max_num_points  : int, optional, default: 1000
        Upper bound on `num_points`. Saves computation time when `num_estimates` is large.

    Returns
    -------
    (alpha, z, L, U) - tuple of scalar and three arrays of size (num_estimates,) containing the confidence level
        as well as the evaluation points, the lower, and the upper confidence bands, respectively.
    """

    # Use shorter variable names to match paper notation
    N = num_estimates
    if num_points is None:
        K = min(N, max_num_points)
    else:
        K = min(num_points, max_num_points)
    M = num_simulations

    # Specify evaluation points
    z = np.linspace(0 + eps, 1 - eps, K)

    # Simulate M samples of size N
    u = np.random.uniform(size=(M, N))

    # Compute minimal coverage probabilities
    gammas = minimal_coverage_probs(z, u)

    # Use insights from paper to compute lower and upper confidence interval
    alpha = 1 - confidence
    gamma = np.percentile(gammas, 100 * alpha)
    lower_ci = scipy_binomial(N, z).ppf(gamma / 2) / N
    upper_ci = scipy_binomial(N, z).ppf(1 - gamma / 2) / N
    return alpha, z, lower_ci, upper_ci



from collections.abc import Sequence, Mapping, Callable

import numpy as np

from ...utils.dict_utils import dicts_to_arrays


def root_mean_squared_error(
    estimates: Mapping[str, np.ndarray] | np.ndarray,
    targets: Mapping[str, np.ndarray] | np.ndarray,
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    normalize: bool = True,
    aggregation: Callable = np.median,
) -> dict[str, any]:
    """
    Computes the (Normalized) Root Mean Squared Error (RMSE/NRMSE) for the given posterior and prior samples.

    Parameters
    ----------
    estimates   : np.ndarray of shape (num_datasets, num_draws_post, num_variables)
        Posterior samples, comprising `num_draws_post` random draws from the posterior distribution
        for each data set from `num_datasets`.
    targets  : np.ndarray of shape (num_datasets, num_variables)
        Prior samples, comprising `num_datasets` ground truths.
    variable_keys : Sequence[str], optional (default = None)
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names : Sequence[str], optional (default = None)
        Optional variable names to show in the output.
    normalize      : bool, optional (default = True)
        Whether to normalize the RMSE using the range of the prior samples.
    aggregation    : callable, optional (default = np.median)
        Function to aggregate the RMSE across draws. Typically `np.mean` or `np.median`.

    Notes
    -----
    Aggregation is performed after computing the RMSE for each posterior draw, instead of first aggregating
    the posterior draws and then computing the RMSE between aggregates and ground truths.

    Returns
    -------
    result : dict
        Dictionary containing:

        - "values" : np.ndarray
            The aggregated (N)RMSE for each variable.
        - "metric_name" : str
            The name of the metric ("RMSE" or "NRMSE").
        - "variable_names" : str
            The (inferred) variable names.
    """

    samples = dicts_to_arrays(
        estimates=estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
    )

    rmse = np.sqrt(np.mean((samples["estimates"] - samples["targets"][:, None, :]) ** 2, axis=0))

    if normalize:
        rmse /= (samples["targets"].max(axis=0) - samples["targets"].min(axis=0))[None, :]
        metric_name = "NRMSE"
    else:
        metric_name = "RMSE"

    rmse = aggregation(rmse, axis=0)
    variable_names = samples["estimates"].variable_names
    return {"values": rmse, "metric_name": metric_name, "variable_names": variable_names}



import keras

from keras.saving import register_keras_serializable as serializable

from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs, fill_triangular_matrix


@serializable(package="bayesflow.links")
class PositiveDefinite(keras.Layer):
    """Activation function to link from flat elements of a lower triangular matrix to a positive definite matrix."""

    def __init__(self, **kwargs):
        super().__init__(**layer_kwargs(kwargs))
        self.built = True

    def call(self, inputs: Tensor) -> Tensor:
        # Build cholesky factor from inputs
        L = fill_triangular_matrix(inputs, positive_diag=True)

        # calculate positive definite matrix from cholesky factors
        psd = keras.ops.matmul(
            L,
            keras.ops.moveaxis(L, -2, -1),  # L transposed
        )
        return psd

    def compute_output_shape(self, input_shape):
        m = input_shape[-1]
        n = int((0.25 + 2.0 * m) ** 0.5 - 0.5)
        return input_shape[:-1] + (n, n)

    def compute_input_shape(self, output_shape):
        """
        Returns the shape of parameterization of a cholesky factor triangular matrix.

        There are m nonzero elements of a lower triangular nxn matrix with m = n * (n + 1) / 2.

        Example
        -------
        >>> PositiveDefinite().compute_output_shape((None, 3, 3))
        6
        """
        n = output_shape[-1]
        m = int(n * (n + 1) / 2)
        return output_shape[:-2] + (m,)



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from keras import ops


def mmd_hypothesis_test(
    mmd_null: np.ndarray,
    mmd_observed: float = None,
    alpha_level: float = 0.05,
    null_color: str | tuple = "#132a70",
    observed_color: str | tuple = "red",
    alpha_color: str | tuple = "orange",
    truncate_v_lines_at_kde: bool = False,
    x_min: float = None,
    x_max: float = None,
    bw_factor: float = 1.5,
):
    """

    Parameters
    ----------
    mmd_null       : np.ndarray
        The samples from the MMD sampling distribution under the null hypothesis "the model is well-specified"
    mmd_observed   : float
        The observed MMD value
    alpha_level    : float, optional, default: 0.05
        The rejection probability (type I error)
    null_color     : str or tuple, optional, default: (0.16407, 0.020171, 0.577478)
        The color of the H0 sampling distribution
    observed_color : str or tuple, optional, default: "red"
        The color of the observed MMD
    alpha_color    : str or tuple, optional, default: "orange"
        The color of the rejection area
    truncate_v_lines_at_kde: bool, optional, default: False
        true: cut off the vlines at the kde
        false: continue kde lines across the plot
    x_min           : float, optional, default: None
        The lower x-axis limit
    x_max           : float, optional, default: None
        The upper x-axis limit
    bw_factor      : float, optional, default: 1.5
        bandwidth (aka. smoothing parameter) of the kernel density estimate

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving
    """

    def draw_v_line_to_kde(x, kde_object, color, label=None, **kwargs):
        kde_x, kde_y = kde_object.lines[0].get_data()
        idx = ops.argmin(ops.abs(kde_x - x))
        plt.vlines(x=x, ymin=0, ymax=kde_y[idx], color=color, linewidth=3, label=label, **kwargs)

    def fill_area_under_kde(kde_object, x_start, x_end=None, **kwargs):
        kde_x, kde_y = kde_object.lines[0].get_data()
        if x_end is not None:
            plt.fill_between(kde_x, kde_y, where=(kde_x >= x_start) & (kde_x <= x_end), interpolate=True, **kwargs)
        else:
            plt.fill_between(kde_x, kde_y, where=(kde_x >= x_start), interpolate=True, **kwargs)

    f = plt.figure(figsize=(8, 4))

    kde = sns.kdeplot(mmd_null, fill=False, linewidth=0, bw_adjust=bw_factor)
    sns.kdeplot(mmd_null, fill=True, alpha=0.12, color=null_color, bw_adjust=bw_factor)

    if truncate_v_lines_at_kde:
        draw_v_line_to_kde(x=mmd_observed, kde_object=kde, color=observed_color, label=r"Observed data")
    else:
        plt.vlines(
            x=mmd_observed,
            ymin=0,
            ymax=plt.gca().get_ylim()[1],
            color=observed_color,
            linewidth=3,
            label=r"Observed data",
        )

    mmd_critical = ops.quantile(mmd_null, 1 - alpha_level)
    fill_area_under_kde(
        kde, mmd_critical, color=alpha_color, alpha=0.5, label=rf"{int(alpha_level * 100)}% rejection area"
    )

    if truncate_v_lines_at_kde:
        draw_v_line_to_kde(x=mmd_critical, kde_object=kde, color=alpha_color)
    else:
        plt.vlines(x=mmd_critical, color=alpha_color, linewidth=3, ymin=0, ymax=plt.gca().get_ylim()[1])

    sns.kdeplot(mmd_null, fill=False, linewidth=3, color=null_color, label=r"$H_0$", bw_adjust=bw_factor)

    plt.xlabel(r"MMD", fontsize=20)
    plt.ylabel("")
    plt.yticks([])
    plt.xlim(x_min, x_max)
    plt.tick_params(axis="both", which="major", labelsize=16)

    plt.legend(fontsize=20)
    sns.despine()

    return f



from functools import singledispatch
import keras


@singledispatch
def find_inference_network(arg, *args, **kwargs):
    raise TypeError(f"Cannot infer inference network from {arg!r}.")


@find_inference_network.register
def _(name: str, *args, **kwargs):
    match name.lower():
        case "coupling_flow":
            from bayesflow.networks import CouplingFlow

            return CouplingFlow(*args, **kwargs)

        case "flow_matching":
            from bayesflow.networks import FlowMatching

            return FlowMatching(*args, **kwargs)

        case "consistency_model":
            from bayesflow.networks import ConsistencyModel

            return ConsistencyModel(*args, **kwargs)

        case unknown_network:
            raise ValueError(f"Unknown inference network: '{unknown_network}'")


@find_inference_network.register
def _(layer: keras.Layer, *args, **kwargs):
    return layer


@find_inference_network.register
def _(model: keras.Model, *args, **kwargs):
    return model



import pytest


@pytest.fixture()
def inference_network():
    from bayesflow.networks import CouplingFlow

    return CouplingFlow(depth=2)


@pytest.fixture()
def random_time_series():
    import keras

    return keras.random.normal(shape=(2, 80, 2))


@pytest.fixture()
def mamba_summary_network():
    from bayesflow.wrappers.mamba import Mamba

    return Mamba(summary_dim=4, feature_dims=(2, 2), state_dims=(4, 4), conv_dims=(8, 8))



import keras

from bayesflow.metrics.functional import maximum_mean_discrepancy
from bayesflow.types import Tensor
from bayesflow.utils import layer_kwargs, find_distribution
from bayesflow.utils.decorators import sanitize_input_shape
from bayesflow.utils.serialization import deserialize


class SummaryNetwork(keras.Layer):
    def __init__(self, base_distribution: str = None, **kwargs):
        super().__init__(**layer_kwargs(kwargs))
        self.base_distribution = find_distribution(base_distribution)

    @sanitize_input_shape
    def build(self, input_shape):
        x = keras.ops.zeros(input_shape)
        z = self.call(x)

        if self.base_distribution is not None:
            self.base_distribution.build(keras.ops.shape(z))

    @sanitize_input_shape
    def compute_output_shape(self, input_shape):
        return keras.ops.shape(self.call(keras.ops.zeros(input_shape)))

    def call(self, x: Tensor, **kwargs) -> Tensor:
        """
        :param x: Tensor of shape (batch_size, set_size, input_dim)

        :param kwargs: Additional keyword arguments.

        :return: Tensor of shape (batch_size, output_dim)
        """
        raise NotImplementedError

    def compute_metrics(self, x: Tensor, stage: str = "training", **kwargs) -> dict[str, Tensor]:
        outputs = self(x, training=stage == "training")

        metrics = {"outputs": outputs}

        if self.base_distribution is not None:
            samples = self.base_distribution.sample((keras.ops.shape(x)[0],))
            mmd = maximum_mean_discrepancy(outputs, samples)
            metrics["loss"] = keras.ops.mean(mmd)

            if stage != "training":
                # compute sample-based validation metrics
                for metric in self.metrics:
                    metrics[metric.name] = metric(outputs, samples)

        return metrics

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))



import inspect
from collections.abc import Callable, Mapping, Sequence
from functools import singledispatch
from types import FunctionType

import numpy as np

from .simulator import Simulator


@singledispatch
def make_simulator(arg, *_, **__):
    """
    This is a dispatch function that will accept a list of simulators (callables) returning
    dictionaries with simulated outputs. The outputs of simulators will be passed to following
    simulators if the latter accept keyword arguments associated with the keys of previous outputs.
    """
    raise TypeError(f"Cannot infer simulator from {arg!r}.")


@make_simulator.register
def _(simulator: Simulator):
    return simulator


@make_simulator.register(FunctionType)
def _(fn: Callable, **kwargs):
    from bayesflow.simulators import LambdaSimulator

    return LambdaSimulator(fn, **kwargs)


@make_simulator.register(Sequence)
def _(
    objs: Sequence[FunctionType],
    obj_kwargs: Mapping[str, dict[str, any]] = None,
    meta_fn: Callable[[], dict[str, np.ndarray]] = None,
    **kwargs,
):
    from bayesflow.simulators import LambdaSimulator, SequentialSimulator

    if obj_kwargs is None:
        obj_kwargs = {}

    # sanity check
    detected_names = {obj.__name__ for obj in objs if hasattr(obj, "__name__")}
    given_names = set(obj_kwargs.keys())

    if not given_names.issubset(detected_names):
        unmatched_names = given_names - detected_names
        msg = (
            f"Found at least one key in obj_kwargs that does not have a match in the object sequence:\n"
            f"{list(unmatched_names)!r}"
        )

        if not all(hasattr(obj, "__name__") for obj in objs):
            msg += (
                "\nThis can happen if the matching objects in the sequence do not have a __name__ attribute. "
                "Pass a dictionary instead to specify names explicitly."
            )

        raise ValueError(msg)

    simulators = []

    for obj in objs:
        if hasattr(obj, "__name__"):
            obj_kwargs = obj_kwargs.get(obj.__name__, {})
        else:
            obj_kwargs = {}

        simulators.append(make_simulator(obj, **obj_kwargs))

    if meta_fn is not None:
        if not inspect.signature(meta_fn).parameters:
            original_meta_fn = meta_fn

            def meta_fn(*_, **__):
                return original_meta_fn()

        meta = LambdaSimulator(meta_fn, is_batched=True)

        simulators = [meta, *simulators]

    return SequentialSimulator(simulators, **kwargs)


@make_simulator.register(Mapping)
def _(
    objs: Mapping[str, FunctionType],
    obj_kwargs: Mapping[str, dict[str, any]] = None,
    meta_fn: Callable[[], dict[str, np.ndarray]] = None,
    **kwargs,
):
    from bayesflow.simulators import LambdaSimulator, SequentialSimulator

    if obj_kwargs is None:
        obj_kwargs = {}

    # sanity check
    detected_names = set(objs.keys())
    given_names = set(obj_kwargs.keys())

    if not given_names.issubset(detected_names):
        unmatched_names = given_names - detected_names
        raise ValueError(
            f"Found at least one key in obj_kwargs that does not have a match in the object mapping:\n"
            f"{list(unmatched_names)!r}"
        )

    simulators = []

    for name, obj in objs.items():
        obj_kwargs = obj_kwargs.get(name, {})
        simulators.append(make_simulator(obj, **obj_kwargs))

    if meta_fn is not None:
        meta = LambdaSimulator(meta_fn, is_batched=True)
        simulators = [meta, *simulators]

    return SequentialSimulator(simulators, **kwargs)



import keras
import numpy as np
import pytest

from bayesflow.utils.serialization import serialize, deserialize

from tests.utils import assert_allclose, assert_layers_equal


def test_build(inference_network, random_samples, random_conditions):
    assert inference_network.built is False

    samples_shape = keras.ops.shape(random_samples)
    conditions_shape = keras.ops.shape(random_conditions) if random_conditions is not None else None

    inference_network.build(samples_shape, conditions_shape=conditions_shape)

    assert inference_network.built is True

    # check the model has variables
    assert inference_network.variables, "Model has no variables."


def test_variable_batch_size(inference_network, random_samples, random_conditions):
    # build with one batch size
    samples_shape = keras.ops.shape(random_samples)
    conditions_shape = keras.ops.shape(random_conditions) if random_conditions is not None else None
    inference_network.build(samples_shape, conditions_shape=conditions_shape)

    # run with another batch size
    batch_sizes = np.random.choice(10, replace=False, size=3)
    for bs in batch_sizes:
        new_input = keras.ops.zeros((bs,) + keras.ops.shape(random_samples)[1:])
        if random_conditions is None:
            new_conditions = None
        else:
            new_conditions = keras.ops.zeros((bs,) + keras.ops.shape(random_conditions)[1:])

        try:
            inference_network(new_input, conditions=new_conditions)
        except NotImplementedError:
            # network is not invertible
            pass
        inference_network(new_input, conditions=new_conditions, inverse=True)


@pytest.mark.parametrize("density", [True, False])
def test_output_structure(density, generative_inference_network, random_samples, random_conditions):
    try:
        output = generative_inference_network(random_samples, conditions=random_conditions, density=density)
    except NotImplementedError:
        # network not invertible
        return

    if density:
        assert isinstance(output, tuple)
        assert len(output) == 2

        forward_output, forward_log_density = output

        assert keras.ops.is_tensor(forward_output)
        assert keras.ops.is_tensor(forward_log_density)
    else:
        assert keras.ops.is_tensor(output)


def test_output_shape(generative_inference_network, random_samples, random_conditions):
    try:
        forward_output, forward_log_density = generative_inference_network(
            random_samples, conditions=random_conditions, density=True
        )
    except NotImplementedError:
        # network is not invertible, not forward function available
        return

    assert keras.ops.shape(forward_output) == keras.ops.shape(random_samples)
    assert keras.ops.shape(forward_log_density) == (keras.ops.shape(random_samples)[0],)

    inverse_output, inverse_log_density = generative_inference_network(
        random_samples, conditions=random_conditions, density=True, inverse=True
    )

    assert keras.ops.shape(inverse_output) == keras.ops.shape(random_samples)
    assert keras.ops.shape(inverse_log_density) == (keras.ops.shape(random_samples)[0],)


def test_cycle_consistency(generative_inference_network, random_samples, random_conditions):
    # cycle-consistency means the forward and inverse methods are inverses of each other
    try:
        forward_output, forward_log_density = generative_inference_network(
            random_samples, conditions=random_conditions, density=True
        )
    except NotImplementedError:
        # network is not invertible, cycle consistency cannot be tested.
        return
    inverse_output, inverse_log_density = generative_inference_network(
        forward_output, conditions=random_conditions, density=True, inverse=True
    )

    assert_allclose(random_samples, inverse_output, atol=1e-3, rtol=1e-3)
    assert_allclose(forward_log_density, inverse_log_density, atol=1e-3, rtol=1e-3)


def test_density_numerically(generative_inference_network, random_samples, random_conditions):
    from bayesflow.utils import jacobian

    try:
        output, log_density = generative_inference_network(random_samples, conditions=random_conditions, density=True)
    except NotImplementedError:
        # network does not support density estimation
        return

    def f(x):
        return generative_inference_network(x, conditions=random_conditions)

    numerical_output, numerical_jacobian = jacobian(f, random_samples, return_output=True)

    # output should be identical, otherwise this test does not work (e.g. for stochastic networks)
    assert_allclose(output, numerical_output)

    log_prob = generative_inference_network.base_distribution.log_prob(output)

    # use change of variables to compute the numerical log density
    numerical_log_density = log_prob + keras.ops.log(keras.ops.abs(keras.ops.det(numerical_jacobian)))

    # use a high tolerance because the numerical jacobian is not very accurate
    assert_allclose(log_density, numerical_log_density, rtol=1e-3, atol=1e-3)


def test_serialize_deserialize(inference_network, random_samples, random_conditions):
    # to save, the model must be built
    xz_shape = keras.ops.shape(random_samples)
    conditions_shape = keras.ops.shape(random_conditions) if random_conditions is not None else None
    inference_network.build(xz_shape, conditions_shape)

    serialized = serialize(inference_network)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)

    assert keras.tree.lists_to_tuples(serialized) == keras.tree.lists_to_tuples(reserialized)


def test_save_and_load(tmp_path, inference_network, random_samples, random_conditions):
    # to save, the model must be built
    xz_shape = keras.ops.shape(random_samples)
    conditions_shape = keras.ops.shape(random_conditions) if random_conditions is not None else None
    inference_network.build(xz_shape, conditions_shape)

    keras.saving.save_model(inference_network, tmp_path / "model.keras")
    loaded = keras.saving.load_model(tmp_path / "model.keras")

    assert_layers_equal(inference_network, loaded)



import numpy as np

from .benchmark_simulator import BenchmarkSimulator


class SLCP(BenchmarkSimulator):
    def __init__(
        self,
        lower_bound: float = -3.0,
        upper_bound: float = 3.0,
        n_obs: int = 4,
        flatten: bool = True,
        rng: np.random.Generator = None,
    ):
        """SLCP simulated benchmark
        See https://arxiv.org/pdf/2101.04653.pdf, Task T.3

        Parameters
        ----------
        lower_bound: float, optional, default: -3.0
            The lower bound of the uniform prior.
        upper_bound: float, optional, default: 3.0
            The upper bound of the uniform prior.
        n_obs: int, optional, default: 4
            The number of observations to generate from the slcp likelihood.
        flatten: bool, optional, default: True
            A flag to indicate whather a 1D (`flatten=True`) or a 2D (`flatten=False`)
            representation of the simulated data is returned.
        rng: np.random.Generator or None, optional, default: None
            An optional random number generator to use.
        """

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.n_obs = n_obs
        self.flatten = flatten
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

    def prior(self):
        """Generates a random draw from a 5-dimensional uniform prior bounded between
        `lower_bound` and `upper_bound` which represents the 5 parameters of the SLCP
        simulator.

        Returns
        -------
        params : np.ndarray of shape (5, )
            A single draw from the 5-dimensional uniform prior.
        """

        return self.rng.uniform(low=self.lower_bound, high=self.upper_bound, size=5)

    def observation_model(self, params: np.ndarray):
        """Generates data from the SLCP model designed as a benchmark for a simple likelihood
        and a complex posterior due to a non-linear pushforward params -> x.

        Parameters
        ----------
        params  : np.ndarray of shape (params, D)
            The location parameters of the Gaussian likelihood.

        Returns
        -------
        x : np.ndarray of shape (n_obs*2, ) or (n_obs, 2), as indictated by the `flatten`
            boolean flag. The sample of simulated data from the SLCP model.
        """

        # Specify 2D location
        loc = np.array([params[0], params[1]])

        # Specify 2D covariance matrix
        s1 = params[2] ** 2
        s2 = params[3] ** 2
        rho = np.tanh(params[4])
        cov = rho * s1 * s2
        S_param = np.array([[s1**2, cov], [cov, s2**2]])

        # Obtain given number of draws from the MVN likelihood
        x = self.rng.multivariate_normal(loc, S_param, size=self.n_obs)
        if self.flatten:
            return x.flatten()
        return x



import numpy as np

from bayesflow.utils.serialization import serializable, serialize

from bayesflow.utils.numpy_utils import one_hot
from .elementwise_transform import ElementwiseTransform


@serializable
class OneHot(ElementwiseTransform):
    """
    Changes data to be one-hot encoded.

    Parameters
    ----------
    num_classes : int
        Number of classes for the encoding.
    """

    def __init__(self, num_classes: int):
        super().__init__()
        self.num_classes = num_classes

    def get_config(self) -> dict:
        return serialize({"num_classes": self.num_classes})

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return one_hot(data, self.num_classes)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.argmax(data, axis=-1)



from functools import partial
import keras

from bayesflow.utils.serialization import serializable
from .functional import root_mean_squared_error


@serializable
class RootMeanSquaredError(keras.metrics.MeanMetricWrapper):
    def __init__(self, name="root_mean_squared_error", dtype=None, **kwargs):
        fn = partial(root_mean_squared_error, **kwargs)
        super().__init__(fn, name=name, dtype=dtype)



import pytest

import keras

from bayesflow.utils.serialization import serializable


@pytest.fixture(params=["coupling_flow", "flow_matching"])
def inference_network(request):
    if request.param == "coupling_flow":
        from bayesflow.networks import CouplingFlow

        return CouplingFlow(depth=2)

    elif request.param == "flow_matching":
        from bayesflow.networks import FlowMatching

        return FlowMatching(subnet_kwargs=dict(widths=(32, 32)), use_optimal_transport=False)


@pytest.fixture(params=["time_series_transformer", "fusion_transformer", "time_series_network", "custom"])
def summary_network(request):
    if request.param == "time_series_transformer":
        from bayesflow.networks import TimeSeriesTransformer

        return TimeSeriesTransformer(embed_dims=(8, 8), mlp_widths=(16, 8), mlp_depths=(1, 1))

    elif request.param == "fusion_transformer":
        from bayesflow.networks import FusionTransformer

        return FusionTransformer(
            embed_dims=(8, 8), mlp_widths=(8, 16), mlp_depths=(2, 1), template_dim=8, bidirectional=False
        )

    elif request.param == "time_series_network":
        from bayesflow.networks import TimeSeriesNetwork

        return TimeSeriesNetwork(filters=4, skip_steps=2)

    elif request.param == "custom":
        from bayesflow.networks import SummaryNetwork

        @serializable
        class Custom(SummaryNetwork):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                self.inner = keras.Sequential([keras.layers.LSTM(8), keras.layers.Dense(4)])

            def call(self, x, **kwargs):
                return self.inner(x, training=kwargs.get("stage") == "training")

        return Custom()



"""
Provides :py:class:`~bayesflow.workflows.BasicWorkflow`, a high-level interface for working on typical BayesFlow
applications without having to worry about the internals.
"""

from .basic_workflow import BasicWorkflow

from ..utils._docs import _add_imports_to_all

_add_imports_to_all(include_modules=[])



import pytest

import keras


@pytest.fixture(params=[2])
def batch_size(request):
    return request.param


@pytest.fixture(params=[2, 3])
def num_features(request):
    return request.param


@pytest.fixture()
def random_samples(batch_size, num_features):
    return keras.random.normal((batch_size, num_features))


@pytest.fixture()
def diagonal_normal():
    from bayesflow.distributions import DiagonalNormal

    return DiagonalNormal()


@pytest.fixture()
def diagonal_student_t():
    from bayesflow.distributions import DiagonalStudentT

    return DiagonalStudentT(df=10)


@pytest.fixture()
def mixture():
    from bayesflow.distributions import DiagonalNormal, DiagonalStudentT, Mixture

    return Mixture([DiagonalNormal(), DiagonalStudentT(df=25)])


@pytest.fixture(params=["diagonal_normal", "diagonal_student_t", "mixture"])
def distribution(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def shape(batch_size, num_features):
    return batch_size, num_features



import keras

from bayesflow.types import Tensor
from bayesflow.utils import issue_url

from .kernels import gaussian, inverse_multiquadratic


def maximum_mean_discrepancy(
    x: Tensor, y: Tensor, kernel: str = "inverse_multiquadratic", unbiased: bool = False, **kwargs
) -> Tensor:
    """Computes a mixture of Gaussian radial basis functions (RBFs) between the samples of x and y.

    See the original paper below for details and different estimators:

    Gretton, A., Borgwardt, K. M., Rasch, M. J., Schölkopf, B., & Smola, A. (2012).
    A kernel two-sample test. The Journal of Machine Learning Research, 13(1), 723-773.
    https://jmlr.csail.mit.edu/papers/v13/gretton12a.html

    Parameters
    ----------
    x        :  Tensor of shape (num_draws_x, num_features)
        Comprises `num_draws_x` Random draws from the "source" distribution `P`.
    y        :  Tensor of shape (num_draws_y, num_features)
        Comprises `num_draws_y` Random draws from the "source" distribution `Q`.
    kernel   : str, optional (default - "inverse_multiquadratic")
        The (mixture of) kernels to be used for the MMD computation.
    unbiased : bool, optional (default - False)
        Whether to use the unbiased MMD estimator. Default is False.

    Returns
    -------
    mmd  : Tensor of shape (1, )
        The biased or unbiased empirical maximum mean discrepancy (MMD) estimator.
    """

    if kernel == "gaussian":
        kernel_fn = gaussian
    elif kernel == "inverse_multiquadratic":
        kernel_fn = inverse_multiquadratic
    else:
        raise ValueError(
            "For now, we only support a gaussian and an inverse_multiquadratic kernel."
            f"If you need a different kernel, please open an issue at {issue_url}"
        )

    if keras.ops.shape(x)[1:] != keras.ops.shape(y)[1:]:
        raise ValueError(
            f"Expected x and y to live in the same feature space, "
            f"but got {keras.ops.shape(x)[1:]} != {keras.ops.shape(y)[1:]}."
        )

    if unbiased:
        m, n = keras.ops.shape(x)[0], keras.ops.shape(y)[0]
        xx = (1.0 / (m * (m + 1))) * keras.ops.sum(kernel_fn(x, x, **kwargs))
        yy = (1.0 / (n * (n + 1))) * keras.ops.sum(kernel_fn(y, y, **kwargs))
        xy = (2.0 / (m * n)) * keras.ops.sum(kernel_fn(x, y, **kwargs))
    else:
        xx = keras.ops.mean(kernel_fn(x, x, **kwargs))
        yy = keras.ops.mean(kernel_fn(y, y, **kwargs))
        xy = keras.ops.mean(kernel_fn(x, y, **kwargs))

    return xx + yy - 2.0 * xy



import keras
from keras.saving import (
    deserialize_keras_object as deserialize,
    serialize_keras_object as serialize,
)
from tests.utils import assert_layers_equal
import pytest


def test_output_structure(point_inference_network, random_samples, random_conditions):
    output = point_inference_network(random_samples, conditions=random_conditions)

    assert isinstance(output, dict)
    for score_key, score in point_inference_network.scores.items():
        head_shapes = score.get_head_shapes_from_target_shape(random_samples.shape)
        assert isinstance(head_shapes, dict)

        for head_key, head_shape in head_shapes.items():
            head_output = output[score_key][head_key]
            assert keras.ops.is_tensor(head_output)
            assert head_output.shape[1:] == head_shape


def test_serialize_deserialize(point_inference_network, random_samples, random_conditions):
    # to save, the model must be built
    point_inference_network(random_samples, conditions=random_conditions)

    serialized = serialize(point_inference_network)
    deserialized = deserialize(serialized)
    reserialized = serialize(deserialized)

    assert serialized == reserialized


def test_save_and_load(tmp_path, point_inference_network, random_samples, random_conditions):
    # to save, the model must be built
    out1 = point_inference_network(random_samples, conditions=random_conditions)

    keras.saving.save_model(point_inference_network, tmp_path / "model.keras")
    loaded = keras.saving.load_model(tmp_path / "model.keras")
    out2 = loaded(random_samples, conditions=random_conditions)

    assert_layers_equal(point_inference_network, loaded)

    for key_outer in out1.keys():
        for key_inner in out1[key_outer].keys():
            assert keras.ops.all(keras.ops.isclose(out1[key_outer][key_inner], out2[key_outer][key_inner])), (
                "Output of original and loaded model differs significantly."
            )


def test_copy_unequal(point_inference_network, random_samples, random_conditions):
    # to save, the model must be built
    point_inference_network(random_samples, conditions=random_conditions)

    copied = keras.models.clone_model(point_inference_network)

    with pytest.raises(AssertionError) as excinfo:
        assert_layers_equal(point_inference_network, copied)

    assert "not equal" in str(excinfo)


def test_save_and_load_quantile(tmp_path, quantile_point_inference_network, random_samples, random_conditions):
    """Test of all nested attributes for a point inference network with a quantile head"""

    # to save, the model must be built
    net = quantile_point_inference_network
    net(random_samples, conditions=random_conditions)

    keras.saving.save_model(net, tmp_path / "model.keras")
    loaded = keras.saving.load_model(tmp_path / "model.keras")

    print(net.get_config())
    assert net.get_config() == loaded.get_config()

    assert_layers_equal(net, loaded)

    for score_key, score in net.scores.items():
        for head_key, head in net.heads[score_key].items():
            net_head = net.heads[score_key][head_key]
            loaded_head = loaded.heads[score_key][head_key]

            net_score = net.scores[score_key]
            loaded_score = loaded.scores[score_key]

            assert keras.ops.all(keras.ops.isclose(net_score._q, loaded_score._q))
            assert keras.ops.all(keras.ops.isclose(net_head.layers[-1].q, loaded_head.layers[-1].q))
            assert keras.ops.all(net_head.layers[-1].anchor_index == loaded_head.layers[-1].anchor_index)

            print(f"Asserting original and serialized and deserialized at heads[{score_key}][{head_key}] to be equal")
            assert_layers_equal(net_head, loaded_head)



from .calibration_error import calibration_error
from .posterior_contraction import posterior_contraction
from .root_mean_squared_error import root_mean_squared_error
from .expected_calibration_error import expected_calibration_error
from .classifier_two_sample_test import classifier_two_sample_test



import keras
from keras import ops

from bayesflow.types import Tensor


def root_mean_squared_error(x1: Tensor, x2: Tensor, normalize: bool = False, **kwargs) -> Tensor:
    """Computes the (normalized) root mean squared error between samples x1 and x2.

    :param x1: Tensor of shape (n, ...)

    :param x2: Tensor of shape (n, ...)

    :param normalize: Normalize the RMSE?

    :param kwargs: Currently ignored

    :return: Tensor of shape (n,)
        The RMSE between x1 and x2 over all remaining dimensions.
    """

    # cannot check first (batch) dimension since it will be unknown at compile time
    if keras.ops.shape(x1)[1:] != keras.ops.shape(x2)[1:]:
        raise ValueError(
            f"Expected x1 and x2 to have the same dimensions, "
            f"but got {keras.ops.shape(x1)[1:]} != {keras.ops.shape(x2)[1:]}."
        )

    # use flattened versions
    x1 = keras.ops.reshape(x1, (keras.ops.shape(x1)[0], -1))
    x2 = keras.ops.reshape(x2, (keras.ops.shape(x2)[0], -1))

    # TODO: how to normalize the RMSE?
    return ops.sqrt(ops.mean(ops.square(x1 - x2), axis=1))



import bayesflow as bf
import pytest


def num_variables(x: dict):
    return sum(arr.shape[-1] for arr in x.values())


def test_backend():
    import matplotlib.pyplot as plt

    # if the local testing backend is not Agg
    # then you may run into issues once you run workflow tests
    # on GitHub, since these use the Agg backend
    assert plt.get_backend() == "Agg"


def test_calibration_ecdf(random_estimates, random_targets, var_names):
    # basic functionality: automatic variable names
    out = bf.diagnostics.plots.calibration_ecdf(random_estimates, random_targets)
    assert len(out.axes) == num_variables(random_estimates)
    assert out.axes[1].title._text == "beta_1"

    # custom variable names
    out = bf.diagnostics.plots.calibration_ecdf(
        estimates=random_estimates,
        targets=random_targets,
        variable_names=var_names,
    )
    assert len(out.axes) == num_variables(random_estimates)
    assert out.axes[1].title._text == "$\\beta_1$"

    # subset of keys with a single scalar key
    out = bf.diagnostics.plots.calibration_ecdf(
        estimates=random_estimates, targets=random_targets, variable_keys="sigma"
    )
    assert len(out.axes) == random_estimates["sigma"].shape[-1]
    assert out.axes[0].title._text == "sigma"

    # use single array instead of dict of arrays as input
    out = bf.diagnostics.plots.calibration_ecdf(
        estimates=random_estimates["beta"],
        targets=random_targets["beta"],
    )
    assert len(out.axes) == random_estimates["beta"].shape[-1]
    # cannot infer the variable names from an array so default names are used
    assert out.axes[1].title._text == "v_1"


def test_calibration_histogram(random_estimates, random_targets):
    # basic functionality: automatic variable names
    out = bf.diagnostics.plots.calibration_histogram(random_estimates, random_targets)
    assert len(out.axes) == num_variables(random_estimates)
    assert out.axes[0].title._text == "beta_0"


def test_loss(history):
    out = bf.diagnostics.loss(history)
    assert len(out.axes) == 1
    assert out.axes[0].title._text == "Loss Trajectory"


def test_recovery(random_estimates, random_targets):
    # basic functionality: automatic variable names
    out = bf.diagnostics.plots.recovery(random_estimates, random_targets)
    assert len(out.axes) == num_variables(random_estimates)
    assert out.axes[2].title._text == "sigma"


def test_z_score_contraction(random_estimates, random_targets):
    # basic functionality: automatic variable names
    out = bf.diagnostics.plots.z_score_contraction(random_estimates, random_targets)
    assert len(out.axes) == num_variables(random_estimates)
    assert out.axes[1].title._text == "beta_1"


def test_pairs_samples(random_priors):
    out = bf.diagnostics.plots.pairs_samples(
        samples=random_priors,
        variable_keys=["beta", "sigma"],
    )
    num_vars = random_priors["sigma"].shape[-1] + random_priors["beta"].shape[-1]
    assert out.axes.shape == (num_vars, num_vars)
    assert out.axes[0, 0].get_ylabel() == "beta_0"
    assert out.axes[2, 2].get_xlabel() == "sigma"


def test_pairs_posterior(random_estimates, random_targets, random_priors):
    # basic functionality: automatic variable names
    out = bf.diagnostics.plots.pairs_posterior(
        random_estimates,
        random_targets,
        dataset_id=1,
    )
    num_vars = num_variables(random_estimates)
    assert out.axes.shape == (num_vars, num_vars)
    assert out.axes[0, 0].get_ylabel() == "beta_0"
    assert out.axes[2, 2].get_xlabel() == "sigma"

    # also plot priors
    out = bf.diagnostics.plots.pairs_posterior(
        estimates=random_estimates,
        targets=random_targets,
        priors=random_priors,
        dataset_id=1,
    )
    num_vars = num_variables(random_estimates)
    assert out.axes.shape == (num_vars, num_vars)
    assert out.axes[0, 0].get_ylabel() == "beta_0"
    assert out.axes[2, 2].get_xlabel() == "sigma"
    assert out.figure.legends[0].get_texts()[0]._text == "Prior"

    with pytest.raises(ValueError):
        bf.diagnostics.plots.pairs_posterior(
            estimates=random_estimates,
            targets=random_targets,
            priors=random_priors,
            dataset_id=[1, 3],
        )


def test_mc_calibration(pred_models, true_models, model_names):
    out = bf.diagnostics.plots.mc_calibration(pred_models, true_models, model_names=model_names)
    assert len(out.axes) == pred_models.shape[-1]
    assert out.axes[0].get_ylabel() == "True Probability"
    assert out.axes[0].get_xlabel() == "Predicted Probability"
    assert out.axes[-1].get_title() == r"$\mathcal{M}_2$"


def test_mc_confusion_matrix(pred_models, true_models, model_names):
    out = bf.diagnostics.plots.mc_confusion_matrix(pred_models, true_models, model_names, normalize="true")
    assert out.axes[0].get_ylabel() == "True model"
    assert out.axes[0].get_xlabel() == "Predicted model"
    assert out.axes[0].get_title() == "Confusion Matrix"



from collections.abc import Mapping, Sequence, Callable

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import keras

from bayesflow.datasets import OnlineDataset, OfflineDataset, DiskDataset
from bayesflow.networks import InferenceNetwork, PointInferenceNetwork, SummaryNetwork
from bayesflow.simulators import Simulator
from bayesflow.adapters import Adapter
from bayesflow.approximators import ContinuousApproximator, PointApproximator
from bayesflow.types import Shape
from bayesflow.utils import find_inference_network, find_summary_network, logging
from bayesflow.diagnostics import metrics as bf_metrics
from bayesflow.diagnostics import plots as bf_plots

from .workflow import Workflow


class BasicWorkflow(Workflow):
    def __init__(
        self,
        simulator: Simulator = None,
        adapter: Adapter = None,
        inference_network: InferenceNetwork | str = "coupling_flow",
        summary_network: SummaryNetwork | str = None,
        initial_learning_rate: float = 5e-4,
        optimizer: keras.optimizers.Optimizer | type = None,
        checkpoint_filepath: str = None,
        checkpoint_name: str = "model",
        save_weights_only: bool = False,
        save_best_only: bool = False,
        inference_variables: Sequence[str] | str = None,
        inference_conditions: Sequence[str] | str = None,
        summary_variables: Sequence[str] | str = None,
        standardize: Sequence[str] | str = "inference_variables",
        **kwargs,
    ):
        """
        This class provides methods to set up, simulate, and fit and validate models using
        amortized Bayesian inference. It allows for both online and offline amortized workflows.

        Parameters
        ----------
        simulator : Simulator, optional
            A Simulator object to generate synthetic data for inference (default is None).
        adapter : Adapter, optional
            Adapter for data processing. If not provided, a default adapter will be used (default is None), but
            you need to make sure to provide the correct names for `inference_variables` and/or `inference_conditions`
            and/or `summary_variables`.
        inference_network : InferenceNetwork or str, optional
            The inference network used for posterior approximation, specified as an instance or by
            name (default is "coupling_flow").
        summary_network : SummaryNetwork or str, optional
            The summary network used for data summarization, specified as an instance or by name (default is None).
        initial_learning_rate : float, optional
            Initial learning rate for the optimizer (default is 5e-4).
        optimizer : type, optional
            The optimizer to be used for training. If None, a default Adam optimizer will be selected (default is None).
        checkpoint_filepath : str, optional
            Directory path where model checkpoints will be saved (default is None).
        checkpoint_name : str, optional
            Name of the checkpoint file (default is "model").
        save_weights_only : bool, optional
            If True, only the model weights will be saved during checkpointing (default is False).
        save_best_only: bool, optional
            If only the best model according to the quantity monitored (loss or validation) at the end of
            each epoch will be saved instead of the last model (default is False). Use with caution,
            as some losses (e.g. flow matching) do not reliably reflect model performance, and outliers in the
            validation data can cause unwanted effects.
        inference_variables : Sequence[str] or str, optional
            Variables for inference as a sequence of strings or a single string (default is None).
            Important for automating diagnostics!
        inference_conditions : Sequence[str] or str, optional
            Variables used as conditions for inference (default is None).
        summary_variables : Sequence[str] or str, optional
            Variables for summarizing data, if any (default is None).
        standardize : Sequence[str] or str, optional
            Variables to standardize during preprocessing (default is "inference_variables").
        **kwargs : dict, optional
            Additional arguments for configuring networks, adapters, optimizers, etc.
        """

        self.inference_network = find_inference_network(inference_network, **kwargs.get("inference_kwargs", {}))

        if summary_network is not None:
            self.summary_network = find_summary_network(summary_network, **kwargs.get("summary_kwargs", {}))
        else:
            self.summary_network = None

        self.simulator = simulator

        adapter = adapter or BasicWorkflow.default_adapter(
            inference_variables, inference_conditions, summary_variables, standardize
        )

        if isinstance(self.inference_network, PointInferenceNetwork):
            constructor = PointApproximator
        else:
            constructor = ContinuousApproximator

        self.approximator = constructor(
            inference_network=self.inference_network, summary_network=self.summary_network, adapter=adapter
        )

        self.initial_learning_rate = initial_learning_rate
        if isinstance(optimizer, type):
            self.optimizer = optimizer(initial_learning_rate, **kwargs.get("optimizer_kwargs", {}))
        else:
            self.optimizer = optimizer

        self.checkpoint_filepath = checkpoint_filepath
        self.checkpoint_name = checkpoint_name
        self.save_weights_only = save_weights_only
        self.save_best_only = save_best_only
        if self.checkpoint_filepath is not None:
            if self.save_weights_only:
                file_ext = self.checkpoint_name + ".weights.h5"
            else:
                file_ext = self.checkpoint_name + ".keras"
            checkpoint_full_filepath = os.path.join(self.checkpoint_filepath, file_ext)
            if os.path.exists(checkpoint_full_filepath):
                msg = (
                    f"Checkpoint file exists: '{checkpoint_full_filepath}'.\n"
                    "Existing checkpoints can _not_ be restored/loaded using this workflow. "
                    "Upon refitting, the checkpoints will be overwritten."
                )
                if not self.save_weights_only:
                    msg += (
                        " To load the stored approximator from the checkpoint, "
                        "use approximator = keras.saving.load_model(...)"
                    )

                logging.warning(msg)
        self.history = None

    @property
    def adapter(self):
        return self.approximator.adapter

    @staticmethod
    def samples_to_data_frame(samples: Mapping[str, np.ndarray]) -> pd.DataFrame:
        """
        Convert a dictionary of samples into a pandas DataFrame.

        Parameters
        ----------
        samples : Mapping[str, np.ndarray]
            A dictionary where keys represent variable names and values are
            arrays containing sampled data.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame where each column corresponds to a variable,
            and rows represent individual samples.
        """
        return pd.DataFrame(keras.tree.map_structure(np.squeeze, samples))

    @staticmethod
    def default_adapter(
        inference_variables: Sequence[str] | str,
        inference_conditions: Sequence[str] | str,
        summary_variables: Sequence[str] | str,
        standardize: Sequence[str] | str,
    ) -> Adapter:
        """
        Create a default adapter for processing inference variables, conditions,
        summaries, and standardization.

        - Converts all float64 values to float32 for computational efficiency.

        Parameters
        ----------
        inference_variables : Sequence[str] or str
            The variables to be treated as inference targets.
        inference_conditions : Sequence[str] or str
            The variables used as conditions for inference.
        summary_variables : Sequence[str] or str
            The variables used for summarization.
        standardize : Sequence[str] or str
            The variables to be standardized.

        Returns
        -------
        Adapter
            A configured Adapter instance that applies dtype conversion,
            concatenation, and optional standardization.
        """

        adapter = (
            Adapter()
            .convert_dtype(from_dtype="float64", to_dtype="float32")
            .concatenate(inference_variables, into="inference_variables")
        )

        if inference_conditions is not None:
            adapter = adapter.concatenate(inference_conditions, into="inference_conditions")
        if summary_variables is not None:
            adapter = adapter.concatenate(summary_variables, into="summary_variables")

        if standardize is not None:
            adapter = adapter.standardize(include=standardize)

        return adapter

    def simulate(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """
        Generates a batch of simulations using the provided simulator.

        Parameters
        ----------
        batch_shape : Shape
            The shape of the batch to be simulated. Typically an integer for simple simulators.
        **kwargs : dict, optional
            Additional keyword arguments passed to the simulator's sample method.

        Returns
        -------
        dict[str, np.ndarray]
            A dictionary where keys represent variable names and values are
            NumPy arrays containing the simulated variables.

        Raises
        ------
        RuntimeError
            If no simulator is provided.
        """
        if self.simulator is not None:
            return self.simulator.sample(batch_shape, **kwargs)
        raise RuntimeError("No simulator provided!")

    def simulate_adapted(self, batch_shape: Shape, **kwargs) -> dict[str, np.ndarray]:
        """
        Generates a batch of simulations and applies the adapter to the result.

        Parameters
        ----------
        batch_shape : Shape
            The shape of the batch to be simulated. Typically an integer for simple simulators.
        **kwargs : dict, optional
            Additional keyword arguments passed to the simulator's sample method.

        Returns
        -------
        dict[str, np.ndarray]
            A dictionary where keys represent variable names and values are
            NumPy arrays containing the adapted simulated variables.

        Raises
        ------
        RuntimeError
            If no simulator is provided.
        """
        if self.simulator is not None:
            return self.adapter(self.simulator.sample(batch_shape, **kwargs))
        raise RuntimeError("No simulator provided!")

    def sample(
        self,
        *,
        num_samples: int,
        conditions: Mapping[str, np.ndarray],
        **kwargs,
    ) -> dict[str, np.ndarray]:
        """
        Draws `num_samples` samples from the approximator given specified conditions.

        Parameters
        ----------
        num_samples : int
            The number of samples to generate.
        conditions : dict[str, np.ndarray]
            A dictionary where keys represent variable names and values are
            NumPy arrays containing the adapted simulated variables. Keys used as summary or inference
            conditions during training should be present.
        **kwargs : dict, optional
            Additional keyword arguments passed to the approximator's sampling function.

        Returns
        -------
        dict[str, np.ndarray]
            A dictionary where keys correspond to variable names and
            values are arrays containing the generated samples.
        """
        return self.approximator.sample(num_samples=num_samples, conditions=conditions, **kwargs)

    def estimate(
        self,
        *,
        conditions: Mapping[str, np.ndarray],
        **kwargs,
    ) -> dict[str, dict[str, np.ndarray | dict[str, np.ndarray]]]:
        """
        Estimates point summaries of inference variables based on specified conditions.

        Parameters
        ----------
        conditions : Mapping[str, np.ndarray]
            A dictionary mapping variable names to arrays representing the conditions for the estimation process.
        **kwargs
            Additional keyword arguments passed to underlying processing functions.

        Returns
        -------
        estimates : dict[str, dict[str, np.ndarray or dict[str, np.ndarray]]]
            The estimates of inference variables in a nested dictionary.

            1. Each first-level key is the name of an inference variable.
            2. Each second-level key is the name of a scoring rule.
            3. (If the scoring rule comprises multiple estimators, each third-level key is the name of an estimator.)

            Each estimator output (i.e., dictionary value that is not itself a dictionary) is an array
            of shape (num_datasets, point_estimate_size, variable_block_size).
        """
        return self.approximator.estimate(conditions=conditions, **kwargs)

    def log_prob(self, data: Mapping[str, np.ndarray], **kwargs) -> np.ndarray:
        """
        Compute the log probability of given variables under the approximator.

        Parameters
        ----------
        data : Mapping[str, np.ndarray]
            A dictionary where keys represent variable names and values are arrays corresponding to the variables'
            realizations.
        **kwargs : dict, optional
            Additional keyword arguments passed to the approximator's log probability function.

        Returns
        -------
        np.ndarray
            An array containing the log probabilities computed from the provided variables.
        """
        return self.approximator.log_prob(data=data, **kwargs)

    def plot_default_diagnostics(
        self,
        test_data: Mapping[str, np.ndarray] | int,
        num_samples: int = 1000,
        variable_keys: Sequence[str] = None,
        variable_names: Sequence[str] = None,
        **kwargs,
    ) -> dict[str, plt.Figure]:
        """
        Generates default diagnostic plots to evaluate the quality of inference. The function produces several
        diagnostic plots, including
        - Loss history (if training history is available).
        - Parameter recovery plots.
        - Calibration ECDF plots.
        - Z-score contraction plots.

        Parameters
        ----------
        test_data : Mapping[str, np.ndarray] or int
            A dictionary containing test data where keys represent variable
            names and values are corresponding data arrays. If an integer
            is provided, that number of test data sets will be generated
            using the simulator (if available).
        num_samples : int, optional
            The number of samples to draw from the approximator for diagnostics,
            by default 1000.
        variable_keys : list or None, optional, default: None
           Select keys from the dictionaries provided in estimates and targets.
           By default, select all keys.
        variable_names : list or None, optional, default: None
            The variable names for nice table plot titles.
        **kwargs : dict, optional
            Additional keyword arguments:

            - `test_data_kwargs`: dict, optional
                Arguments to pass to the simulator when generating test data.
            - `approximator_kwargs`: dict, optional
                Arguments to pass to the approximator's sampling function.
            - `loss_kwargs`: dict, optional
                Arguments for customizing the loss plot.
            - `recovery_kwargs`: dict, optional
                Arguments for customizing the parameter recovery plot.
            - `calibration_ecdf_kwargs`: dict, optional
                Arguments for customizing the empirical cumulative distribution
                function (ECDF) calibration plot.
            - `z_score_contraction_kwargs`: dict, optional
                Arguments for customizing the z-score contraction plot.

        Returns
        -------
        dict[str, plt.Figure]
            A dictionary where keys correspond to different diagnostic plot
            types, and values are the respective matplotlib Figure objects.
        """

        samples, test_data = self._prepare_for_diagnostics(test_data, num_samples, **kwargs)

        figures = dict()

        if self.history is not None:
            figures["losses"] = bf_plots.loss(self.history, **kwargs.get("loss_kwargs", {}))

        plot_fns = {
            "recovery": bf_plots.recovery,
            "calibration_ecdf": bf_plots.calibration_ecdf,
            "z_score_contraction": bf_plots.z_score_contraction,
        }

        for k, plot_fn in plot_fns.items():
            figures[k] = plot_fn(
                estimates=samples,
                targets=test_data,
                variable_keys=variable_keys,
                variable_names=variable_names,
                **kwargs.get(f"{k}_kwargs", {}),
            )

        return figures

    def plot_custom_diagnostics(
        self,
        test_data: Mapping[str, np.ndarray] | int,
        plot_fns: Mapping[str, Callable],
        num_samples: int = 1000,
        variable_keys: Sequence[str] = None,
        variable_names: Sequence[str] = None,
        **kwargs,
    ) -> dict[str, plt.Figure]:
        """
        Generates custom diagnostic plots to evaluate the quality of inference. The functions passed should have
        the following signature:
        - fn(samples, inference_variables, variable_names)

        They should also return a single matplotlib Figure object.

        Parameters
        ----------
        test_data : Mapping[str, np.ndarray] or int
            A dictionary containing test data where keys represent variable
            names and values are corresponding data arrays. If an integer
            is provided, that number of test data sets will be generated
            using the simulator (if available).
        plot_fns: Mapping[str, Callable]
            A dictionary containing custom plotting functions where keys represent
            the function names and values correspond to the functions themselves.
            The functions should have a signature of fn(samples, inference_variables, variable_names)
        num_samples : int, optional
            The number of samples to draw from the approximator for diagnostics,
            by default 1000.
        variable_keys : list or None, optional, default: None
           Select keys from the dictionaries provided in estimates and targets.
           By default, select all keys.
        variable_names : list or None, optional, default: None
            The variable names for nice table plot titles.
        **kwargs : dict, optional
            Additional keyword arguments:

            - `test_data_kwargs`: dict, optional
                Arguments to pass to the simulator when generating test data.
            - `approximator_kwargs`: dict, optional
                Arguments to pass to the approximator's sampling function.
            - `loss_kwargs`: dict, optional
                Arguments for customizing the loss plot.
            - `recovery_kwargs`: dict, optional
                Arguments for customizing the parameter recovery plot.
            - `calibration_ecdf_kwargs`: dict, optional
                Arguments for customizing the empirical cumulative distribution
                function (ECDF) calibration plot.
            - `z_score_contraction_kwargs`: dict, optional
                Arguments for customizing the z-score contraction plot.

        Returns
        -------
        dict[str, plt.Figure]
            A dictionary where keys correspond to different diagnostic plot
            types, and values are the respective matplotlib Figure objects.
        """

        samples, test_data = self._prepare_for_diagnostics(test_data, num_samples, **kwargs)

        figures = dict()
        for key, plot_fn in plot_fns.items():
            figures[key] = plot_fn(samples, test_data, variable_keys=variable_keys, variable_names=variable_names)
        return figures

    def plot_diagnostics(self, **kwargs):
        logging.warning(
            "This function will be deprecated in future versions. Please, use plot_default_diagnostics"
            "or plot_custom_diagnositcs if you want to use your custom diagnostics."
        )
        return self.plot_default_diagnostics(**kwargs)

    def compute_default_diagnostics(
        self,
        test_data: Mapping[str, np.ndarray] | int,
        num_samples: int = 1000,
        variable_keys: Sequence[str] = None,
        variable_names: Sequence[str] = None,
        as_data_frame: bool = True,
        **kwargs,
    ) -> Sequence[dict] | pd.DataFrame:
        """
        Computes default diagnostic metrics to evaluate the quality of inference. The function computes several
        diagnostic metrics, including:
        - Root Mean Squared Error (RMSE)
        - Posterior contraction
        - Calibration error

        Parameters
        ----------
        test_data : Mapping[str, np.ndarray] or int
            A dictionary containing test data where keys represent variable
            names and values are corresponding realizations. If an integer
            is provided, that number of test data sets will be generated
            using the simulator (if available).
        num_samples : int, optional
            The number of samples to draw from the approximator for diagnostics,
            by default 1000.
        variable_keys : list or None, optional, default: None
           Select keys from the dictionaries provided in estimates and targets.
           By default, select all keys.
        variable_names : list or None, optional, default: None
            The parameter names for nice table plot titles.
        as_data_frame : bool, optional
            Whether to return the results as a pandas DataFrame (default: True).
            If False, a sequence of dictionaries with metric values is returned.
        **kwargs : dict, optional
            Additional keyword arguments:

            - `test_data_kwargs`: dict, optional
                Arguments to pass to the simulator when generating test data.
            - `approximator_kwargs`: dict, optional
                Arguments to pass to the approximator's sampling function.
            - `root_mean_squared_error_kwargs`: dict, optional
                Arguments for customizing the RMSE computation.
            - `posterior_contraction_kwargs`: dict, optional
                Arguments for customizing the posterior contraction computation.
            - `calibration_error_kwargs`: dict, optional
                Arguments for customizing the calibration error computation.

        Returns
        -------
        Sequence[dict] or pd.DataFrame
            If `as_data_frame` is True, returns a pandas DataFrame containing
            the computed diagnostic metrics for each variable. Otherwise,
            returns a sequence of dictionaries with metric values.
        """

        samples, test_data = self._prepare_for_diagnostics(test_data, num_samples, **kwargs)

        root_mean_squared_error = bf_metrics.root_mean_squared_error(
            estimates=samples,
            targets=test_data,
            variable_keys=variable_keys,
            variable_names=variable_names,
            **kwargs.get("root_mean_squared_error_kwargs", {}),
        )

        contraction = bf_metrics.posterior_contraction(
            estimates=samples,
            targets=test_data,
            variable_keys=variable_keys,
            variable_names=variable_names,
            **kwargs.get("posterior_contraction_kwargs", {}),
        )

        calibration_errors = bf_metrics.calibration_error(
            estimates=samples,
            targets=test_data,
            variable_keys=variable_keys,
            variable_names=variable_names,
            **kwargs.get("calibration_error_kwargs", {}),
        )

        if as_data_frame:
            metrics = pd.DataFrame(
                {
                    root_mean_squared_error["metric_name"]: root_mean_squared_error["values"],
                    contraction["metric_name"]: contraction["values"],
                    calibration_errors["metric_name"]: calibration_errors["values"],
                },
                index=variable_keys or root_mean_squared_error["variable_names"],
            ).T
        else:
            metrics = (root_mean_squared_error, contraction, calibration_errors)

        return metrics

    def compute_custom_diagnostics(
        self,
        test_data: Mapping[str, np.ndarray] | int,
        metrics: Mapping[str, Callable],
        num_samples: int = 1000,
        variable_keys: Sequence[str] = None,
        variable_names: Sequence[str] = None,
        as_data_frame: bool = True,
        **kwargs,
    ) -> Sequence[Mapping] | pd.DataFrame:
        """
        Computes custom diagnostic metrics to evaluate the quality of inference. The metric functions should
        have a signature of:

        - metric_fn(samples, inference_variables, variable_names, variable_keys) or
        - metric_fn(samples, inference_variables, **kwargs)

        And return a dictionary containing the metric name in 'name' key and the metric values in a 'values' key.

        Parameters
        ----------
        test_data : Mapping[str, np.ndarray] or int
            A dictionary containing test data where keys represent variable
            names and values are corresponding realizations. If an integer
            is provided, that number of test data sets will be generated
            using the simulator (if available).
        metrics: Mapping[str, Callable]
            A dictionary containing custom metric functions where keys represent
            the function names and values correspond to the functions themselves.
            The functions should have a signature of fn(samples, inference_variables, variable_names)
        num_samples : int, optional
            The number of samples to draw from the approximator for diagnostics,
            by default 1000.
        variable_keys : list or None, optional, default: None
           Select keys from the dictionaries provided in estimates and targets.
           By default, select all keys.
        variable_names : list or None, optional, default: None
            The variable names for nice plot titles.
        as_data_frame : bool, optional
            Whether to return the results as a pandas DataFrame (default: True).
            If False, a sequence of dictionaries with metric values is returned.
        **kwargs : dict, optional
            Additional keyword arguments:

            - `test_data_kwargs`: dict, optional
                Arguments to pass to the simulator when generating test data.
            - `approximator_kwargs`: dict, optional
                Arguments to pass to the approximator's sampling function.
            - `root_mean_squared_error_kwargs`: dict, optional
                Arguments for customizing the RMSE computation.
            - `posterior_contraction_kwargs`: dict, optional
                Arguments for customizing the posterior contraction computation.
            - `calibration_error_kwargs`: dict, optional
                Arguments for customizing the calibration error computation.

        Returns
        -------
        Sequence[dict] or pd.DataFrame
            If `as_data_frame` is True, returns a pandas DataFrame containing
            the computed diagnostic metrics for each variable. Otherwise,
            returns a sequence of dictionaries with metric values.
        """

        samples, test_data = self._prepare_for_diagnostics(test_data, num_samples, **kwargs)

        metrics_dict = {}
        for key, metric_fn in metrics.items():
            metric = metric_fn(samples, test_data, variable_keys=variable_keys, variable_names=variable_names)
            metrics_dict[metric["name"]] = metric["values"]

        if as_data_frame:
            return pd.DataFrame(metrics_dict, index=variable_names)
        return metrics_dict

    def compute_diagnostics(self, **kwargs):
        logging.warning(
            "This function will be deprecated in future versions. Please, use plot_default_diagnostics"
            "or compute_custom_diagnositcs if you want to use your own metrics."
        )
        return self.compute_default_diagnostics(**kwargs)

    def fit_offline(
        self,
        data: Mapping[str, np.ndarray],
        epochs: int = 100,
        batch_size: int = 32,
        keep_optimizer: bool = False,
        validation_data: Mapping[str, np.ndarray] | int = None,
        **kwargs,
    ) -> keras.callbacks.History:
        """
        Train the approximator offline using a fixed dataset. This approach will be faster than online training,
        since no computation time is spent in generating new data for each batch, but it assumes that simulations
        can fit in memory.

        Parameters
        ----------
        data : Mapping[str, np.ndarray]
            A dictionary containing training data where keys represent variable
            names and values are corresponding realizations.
        epochs : int, optional
            The number of training epochs, by default 100. Consider increasing this number for free-form inference
            networks, such as FlowMatching or ConsistencyModel, which generally need more epochs than CouplingFlows.
        batch_size : int, optional
            The batch size used for training, by default 32.
        keep_optimizer : bool, optional
            Whether to retain the current state of the optimizer after training,
            by default False.
        validation_data : Mapping[str, np.ndarray] or int, optional
            A dictionary containing validation data. If an integer is provided,
            that number of validation samples will be generated (if supported).
            By default, no validation data is used.
        **kwargs : dict, optional
            Additional keyword arguments passed to the underlying `_fit` method.

        Returns
        -------
        history : keras.callbacks.History
            A history object containing training history, where keys correspond to
            logged metrics (e.g., loss values) and values are arrays tracking
            metric evolution over epochs.
        """

        dataset = OfflineDataset(data=data, batch_size=batch_size, adapter=self.adapter)

        return self._fit(
            dataset, epochs, strategy="online", keep_optimizer=keep_optimizer, validation_data=validation_data, **kwargs
        )

    def fit_online(
        self,
        epochs: int = 100,
        num_batches_per_epoch: int = 100,
        batch_size: int = 32,
        keep_optimizer: bool = False,
        validation_data: Mapping[str, np.ndarray] | int = None,
        **kwargs,
    ) -> keras.callbacks.History:
        """
        Train the approximator using an online data-generating process. The dataset is dynamically generated during
        training, making this approach suitable for scenarios where generating new simulations is computationally cheap.

        Parameters
        ----------
        epochs : int, optional
            The number of training epochs, by default 100.
        num_batches_per_epoch : int, optional
            The number of batches generated per epoch, by default 100.
        batch_size : int, optional
            The batch size used for training, by default 32.
        keep_optimizer : bool, optional
            Whether to retain the current state of the optimizer after training,
            by default False.
        validation_data : Mapping[str, np.ndarray] or int, optional
            A dictionary containing validation data. If an integer is provided,
            that number of validation samples will be generated (if supported).
            By default, no validation data is used.
        **kwargs : dict, optional
            Additional keyword arguments passed to the underlying `_fit` method.

        Returns
        -------
        history : keras.callbacks.History
            A history object containing training history, where keys correspond to
            logged metrics (e.g., loss values) and values are arrays tracking
            metric evolution over epochs.
        """

        dataset = OnlineDataset(
            simulator=self.simulator, batch_size=batch_size, num_batches=num_batches_per_epoch, adapter=self.adapter
        )

        return self._fit(
            dataset, epochs, strategy="online", keep_optimizer=keep_optimizer, validation_data=validation_data, **kwargs
        )

    def fit_disk(
        self,
        root: os.PathLike,
        pattern: str = "*.pkl",
        batch_size: int = 32,
        load_fn: callable = None,
        epochs: int = 100,
        keep_optimizer: bool = False,
        validation_data: Mapping[str, np.ndarray] | int = None,
        **kwargs,
    ) -> keras.callbacks.History:
        """
        Train the approximator using data stored on disk. This approach is suitable for large sets of simulations that
        don't fit in memory.

        Parameters
        ----------
        root : os.PathLike
            The root directory containing the dataset files.
        pattern : str, optional
            A filename pattern to match dataset files, by default ``"*.pkl"``.
        batch_size : int, optional
            The batch size used for training, by default 32.
        load_fn : callable, optional
            A function to load dataset files. If None, a default loading
            function is used.
        epochs : int, optional
            The number of training epochs, by default 100. Consider increasing this number for free-form inference
            networks, such as FlowMatching or ConsistencyModel, which generally need more epochs than CouplingFlows.
        keep_optimizer : bool, optional
            Whether to retain the current state of the optimizer after training,
            by default False.
        validation_data : Mapping[str, np.ndarray] or int, optional
            A dictionary containing validation data. If an integer is provided,
            that number of validation samples will be generated (if supported).
            By default, no validation data is used.
        **kwargs : dict, optional
            Additional keyword arguments passed to the underlying `_fit` method.

        Returns
        -------
        history : keras.callbacks.History
            A history object containing training history, where keys correspond to
            logged metrics (e.g., loss values) and values are arrays tracking
            metric evolution over epochs.
        """

        dataset = DiskDataset(root=root, pattern=pattern, batch_size=batch_size, load_fn=load_fn, adapter=self.adapter)

        return self._fit(
            dataset, epochs, strategy="online", keep_optimizer=keep_optimizer, validation_data=validation_data, **kwargs
        )

    def build_optimizer(self, epochs: int, num_batches: int, strategy: str) -> keras.Optimizer | None:
        """
        Build and initialize the optimizer based on the training strategy. Uses a cosine decay learning rate schedule,
        where the final learning rate is proportional to the square of the initial learning rate, as found to work
        best in SBI.

        The default optimizer will use 5% of the epochs as warmup; during the warmup phase, the learning rate
        will be increased from 10% of the initial learning rate to initial learning rate supplied to the workflow.

        Parameters
        ----------
        epochs : int
            The total number of training epochs.
        num_batches : int
            The number of batches per epoch.
        strategy : str
            The training strategy, either "online" or another mode that
            applies weight decay. For "online" training, an Adam optimizer with gradient clipping is used. For other
            strategies, AdamW is used with weight decay to encourage regularization.

        Returns
        -------
        keras.Optimizer or None
            The initialized optimizer if it was not already set. Returns None
            if the optimizer was already defined.
        """

        if self.optimizer is not None:
            return

        total_steps = int(epochs * num_batches)
        warmup_steps = int(0.05 * epochs * num_batches)
        decay_steps = total_steps - warmup_steps

        # Default case
        learning_rate = keras.optimizers.schedules.CosineDecay(
            initial_learning_rate=0.1 * self.initial_learning_rate,
            warmup_target=self.initial_learning_rate,
            warmup_steps=warmup_steps,
            decay_steps=decay_steps,
            alpha=0,
        )

        # Use adam for online learning, apply weight decay otherwise
        if strategy.lower() == "online":
            self.optimizer = keras.optimizers.Adam(learning_rate, clipnorm=1.5)
        else:
            self.optimizer = keras.optimizers.AdamW(learning_rate, weight_decay=5e-3, clipnorm=1.5)

    def _fit(
        self,
        dataset: keras.utils.PyDataset,
        epochs: int,
        strategy: str,
        keep_optimizer: bool,
        validation_data: Mapping[str, np.ndarray] | int,
        **kwargs,
    ) -> keras.callbacks.History:
        if validation_data is not None:
            if isinstance(validation_data, int) and self.simulator is not None:
                validation_data = self.simulator.sample(validation_data, **kwargs.pop("validation_data_kwargs", {}))
            elif isinstance(validation_data, int):
                raise ValueError(f"No simulator found for generating {validation_data} data sets.")

            validation_data = OfflineDataset(data=validation_data, batch_size=dataset.batch_size, adapter=self.adapter)
            monitor = "val_loss"
        else:
            monitor = "loss"

        if self.checkpoint_filepath is not None:
            if self.save_weights_only:
                file_ext = self.checkpoint_name + ".weights.h5"
            else:
                file_ext = self.checkpoint_name + ".keras"

            model_checkpoint_callback = keras.callbacks.ModelCheckpoint(
                filepath=os.path.join(self.checkpoint_filepath, file_ext),
                monitor=monitor,
                mode="min",
                save_best_only=self.save_best_only,
                save_weights_only=self.save_weights_only,
                save_freq="epoch",
            )

            if kwargs.get("callbacks") is not None:
                kwargs["callbacks"].append(model_checkpoint_callback)
            else:
                kwargs["callbacks"] = [model_checkpoint_callback]

        self.build_optimizer(epochs, dataset.num_batches, strategy=strategy)

        if not self.approximator.built:
            self.approximator.compile(optimizer=self.optimizer, metrics=kwargs.pop("metrics", None))

        try:
            self.history = self.approximator.fit(
                dataset=dataset, epochs=epochs, validation_data=validation_data, **kwargs
            )
            self._on_training_finished()
            return self.history
        finally:
            if not keep_optimizer:
                self.optimizer = None

    def _prepare_for_diagnostics(self, test_data: Mapping[str, np.ndarray] | int, num_samples: int = 1000, **kwargs):
        if isinstance(test_data, int) and self.simulator is not None:
            test_data = self.simulator.sample(test_data, **kwargs.pop("test_data_kwargs", {}))
        elif isinstance(test_data, int):
            raise ValueError(f"No simulator found for generating {test_data} data sets.")

        samples = self.approximator.sample(
            num_samples=num_samples, conditions=test_data, **kwargs.get("approximator_kwargs", {})
        )

        return samples, test_data

    def _on_training_finished(self):
        if self.checkpoint_filepath is not None:
            if self.save_weights_only:
                file_ext = self.checkpoint_name + ".weights.h5"
            else:
                file_ext = self.checkpoint_name + ".keras"

            logging.info(f"""Training is now finished.
            You can find the trained approximator at '{self.checkpoint_filepath}/{self.checkpoint_name}.{file_ext}'.
            To load it, use approximator = keras.saving.load_model(...).""")



import inspect


class monkey_patch:
    """
    Experimental, single-thread, stack-based monkey patching. Use with caution.

    Built for compatibility with recursive functions. Allows extending the functionality and base cases
    of recursive functions without modifying the original source. For example, adding type serialization
    to `keras.saving.deserialize_keras_object`, which acts recursively on collections.

    Usage:

    def replacement_fn():
        with monkey_patch(original_fn, replacement_fn) as original:
            # all calls to `original_fn` are replaced with calls to `replacement_fn`
            # the only exception is calls specifically to `original` in this scope
            return original()

        # original_fn is restored as soon as the scope exits
    """

    stack = []

    def __init__(self, original, replacement):
        if inspect.getmodule(original) == inspect.getmodule(replacement):
            raise ValueError("Cannot monkey patch the same module.")

        self.original = original
        self.replacement = replacement

        self.original_module = inspect.getmodule(original)
        self.original_module_dict = self.original_module.__dict__
        self.original_name = original.__name__

        self.is_redundant = None

    def __enter__(self):
        if self.stack:
            last_original, last_replacement = self.stack[-1]
            if self.original is last_original and self.replacement is last_replacement:
                self.is_redundant = True
                return self.original

        self.is_redundant = False
        self.stack.append((self.original, self.replacement))

        self.original_module_dict[self.original_name] = self.replacement
        return self.original

    def __exit__(self, exc_type, exc_value, traceback):
        if self.is_redundant:
            return

        self.original_module_dict[self.original_name] = self.original
        self.stack.pop()



import numpy as np

from bayesflow.utils.serialization import serializable, serialize
from bayesflow.utils.numpy_utils import (
    inverse_sigmoid,
    inverse_softplus,
    sigmoid,
    softplus,
)

from .elementwise_transform import ElementwiseTransform


@serializable
class Constrain(ElementwiseTransform):
    """
    Constrains neural network predictions of a data variable to specified bounds.

    Parameters
    ----------
    * : str
        String containing the name of the data variable to be transformed e.g. "sigma". See examples below.

    lower : int or float or np.darray, optional
        Lower bound for named data variable.
    upper : int or float or np.darray, optional
        Upper bound for named data variable.
    method : str, optional
        Method by which to shrink the network predictions space to specified bounds. Choose from
        - Double bounded methods: sigmoid, expit, (default = sigmoid)
        - Lower bound only methods: softplus, exp, (default = softplus)
        - Upper bound only methods: softplus, exp, (default = softplus)
    inclusive : {'both', 'lower', 'upper', 'none'}, optional
        Indicates which bounds are inclusive (or exclusive).
        - "both" (default): Both lower and upper bounds are inclusive.
        - "lower": Lower bound is inclusive, upper bound is exclusive.
        - "upper": Lower bound is exclusive, upper bound is inclusive.
        - "none": Both lower and upper bounds are exclusive.
    epsilon : float, optional
        Small value to ensure inclusive bounds are not violated.
        Current default is 1e-15 as this ensures finite outcomes
        with the default transformations applied to data exactly at the boundaries.

    Examples
    --------

    1) Let sigma be the standard deviation of a normal distribution,
    then sigma should always be greater than zero.

    >>> adapter = (
        bf.Adapter()
        .constrain("sigma", lower=0)
    )

    2) Suppose p is the parameter for a binomial distribution where p must be in
    [0,1] then we would constrain the neural network to estimate p in the following way.

    >>> import bayesflow as bf
    >>> adapter = bf.Adapter()
    >>> adapter.constrain("p", lower=0, upper=1, method="sigmoid", inclusive="both")
    """

    def __init__(
        self,
        *,
        lower: int | float | np.ndarray = None,
        upper: int | float | np.ndarray = None,
        method: str = "default",
        inclusive: str = "both",
        epsilon: float = 1e-15,
    ):
        super().__init__()

        if lower is None and upper is None:
            raise ValueError("At least one of 'lower' or 'upper' must be provided.")

        if lower is not None and upper is not None:
            # double bounded case
            if np.any(lower >= upper):
                raise ValueError("The lower bound must be strictly less than the upper bound.")

            match method:
                case "default" | "sigmoid" | "expit" | "logit":

                    def constrain(x):
                        return (upper - lower) * sigmoid(x) + lower

                    def unconstrain(x):
                        return inverse_sigmoid((x - lower) / (upper - lower))
                case str() as name:
                    raise ValueError(f"Unsupported method name for double bounded constraint: '{name}'.")
                case other:
                    raise TypeError(f"Expected a method name, got {other!r}.")
        elif lower is not None:
            # lower bounded case
            match method:
                case "default" | "softplus":

                    def constrain(x):
                        return softplus(x) + lower

                    def unconstrain(x):
                        return inverse_softplus(x - lower)
                case "exp" | "log":

                    def constrain(x):
                        return np.exp(x) + lower

                    def unconstrain(x):
                        return np.log(x - lower)
                case str() as name:
                    raise ValueError(f"Unsupported method name for single bounded constraint: '{name}'.")
                case other:
                    raise TypeError(f"Expected a method name, got {other!r}.")
        else:
            # upper bounded case
            match method:
                case "default" | "softplus":

                    def constrain(x):
                        return -softplus(-x) + upper

                    def unconstrain(x):
                        return -inverse_softplus(-(x - upper))
                case "exp" | "log":

                    def constrain(x):
                        return -np.exp(-x) + upper

                    def unconstrain(x):
                        return -np.log(-x + upper)
                case str() as name:
                    raise ValueError(f"Unsupported method name for single bounded constraint: '{name}'.")
                case other:
                    raise TypeError(f"Expected a method name, got {other!r}.")

        self.lower = lower
        self.upper = upper
        self.method = method
        self.inclusive = inclusive
        self.epsilon = epsilon

        self.constrain = constrain
        self.unconstrain = unconstrain

        # do this last to avoid serialization issues
        match inclusive:
            case "lower":
                if lower is not None:
                    lower = lower - epsilon
            case "upper":
                if upper is not None:
                    upper = upper + epsilon
            case True | "both":
                if lower is not None:
                    lower = lower - epsilon
                if upper is not None:
                    upper = upper + epsilon
            case False | None | "none":
                pass
            case other:
                raise ValueError(f"Unsupported value for 'inclusive': {other!r}.")

    def get_config(self) -> dict:
        config = {
            "lower": self.lower,
            "upper": self.upper,
            "method": self.method,
            "inclusive": self.inclusive,
            "epsilon": self.epsilon,
        }
        return serialize(config)

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        # forward means data space -> network space, so unconstrain the data
        return self.unconstrain(data)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        # inverse means network space -> data space, so constrain the data
        return self.constrain(data)



import numpy as np

from bayesflow.utils.serialization import serializable

from .elementwise_transform import ElementwiseTransform


@serializable
class Sqrt(ElementwiseTransform):
    """Square-root transform a variable.

    Examples
    --------
    >>> adapter = bf.Adapter().sqrt(["x"])
    """

    def forward(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.sqrt(data)

    def inverse(self, data: np.ndarray, **kwargs) -> np.ndarray:
        return np.square(data)

    def get_config(self) -> dict:
        return {}



import copy
import keras
import pytest


from tests.utils import assert_models_equal
from tests.utils import InterruptFitCallback, FitInterruptedError


@pytest.mark.parametrize("jit_compile", [False, True])
def test_compile(approximator, random_samples, jit_compile):
    approximator.compile(jit_compile=jit_compile)


def test_fit(approximator, train_dataset, validation_dataset, batch_size):
    from bayesflow.metrics import MaximumMeanDiscrepancy
    from bayesflow.networks import PointInferenceNetwork

    inference_metrics = []
    if not isinstance(approximator.inference_network, PointInferenceNetwork):
        inference_metrics += [MaximumMeanDiscrepancy()]
    approximator.compile(inference_metrics=inference_metrics)

    mock_data = train_dataset[0]
    mock_data = keras.tree.map_structure(keras.ops.convert_to_tensor, mock_data)
    approximator.build_from_data(mock_data)

    untrained_weights = copy.deepcopy(approximator.weights)
    untrained_metrics = approximator.evaluate(validation_dataset, return_dict=True)

    approximator.fit(dataset=train_dataset, epochs=50, batch_size=batch_size)

    trained_weights = approximator.weights
    trained_metrics = approximator.evaluate(validation_dataset, return_dict=True)

    # check weights have changed during training
    assert any([keras.ops.any(~keras.ops.isclose(u, t)) for u, t in zip(untrained_weights, trained_weights)])

    assert isinstance(untrained_metrics, dict)
    assert isinstance(trained_metrics, dict)

    # test that metrics are improving
    metric_names = ["loss"]
    if not isinstance(approximator.inference_network, PointInferenceNetwork):
        metric_names += ["maximum_mean_discrepancy/inference_maximum_mean_discrepancy"]
    for metric in metric_names:
        assert metric in untrained_metrics
        assert metric in trained_metrics

        # TODO: this is too flaky
        # assert trained_metrics[metric] <= untrained_metrics[metric]

    pytest.skip("Marking as skipped because metrics are currently untested.")


def test_serialize_deserialize(tmp_path, approximator, train_dataset):
    mock_data = train_dataset[0]
    mock_data = keras.tree.map_structure(keras.ops.convert_to_tensor, mock_data)
    approximator.build_from_data(mock_data)

    keras.saving.save_model(approximator, tmp_path / "model.keras")
    loaded_approximator = keras.saving.load_model(tmp_path / "model.keras")

    assert_models_equal(approximator, loaded_approximator)


@pytest.mark.skip(reason="not implemented")
def test_interrupt_and_resume_fit(tmp_path, amortizer, dataset):
    # TODO: test the InterruptFitCallback
    amortizer.compile(optimizer="AdamW")

    callbacks = [
        InterruptFitCallback(epochs=1, error_type=FitInterruptedError),
        keras.callbacks.ModelCheckpoint(tmp_path / "model.keras"),
    ]

    with pytest.raises(FitInterruptedError):
        # interrupted fit (epochs 0-1)
        amortizer.fit(dataset, epochs=2, callbacks=callbacks)

    assert (tmp_path / "model.keras").exists(), "checkpoint has not been created"

    loaded_amortizer = keras.saving.load_model(tmp_path / "model.keras")

    # TODO: verify the fit is actually resumed (and not just started new with the existing weights)
    #  might require test code change
    # resume fit (epochs 1-2)
    loaded_amortizer.fit(dataset, epochs=2)



from collections.abc import Mapping

import numpy as np

import keras

from bayesflow.adapters import Adapter
from bayesflow.utils import logging


class OfflineDataset(keras.utils.PyDataset):
    """
    A dataset that is pre-simulated and stored in memory. When storing and loading data from disk, it is recommended to
    save any pre-simulated data in raw form and create the `OfflineDataset` object only after loading in the raw data.
    See the `DiskDataset` class for handling large datasets that are split into multiple smaller files.
    """

    def __init__(
        self,
        data: Mapping[str, np.ndarray],
        batch_size: int,
        adapter: Adapter | None,
        num_samples: int = None,
        *,
        stage: str = "training",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.batch_size = batch_size
        self.data = data
        self.adapter = adapter
        self.stage = stage

        if num_samples is None:
            self.num_samples = self._get_num_samples_from_data(data)
            logging.debug(f"Automatically determined {self.num_samples} samples in data.")
        else:
            self.num_samples = num_samples

        self.indices = np.arange(self.num_samples, dtype="int64")

        self.shuffle()

    def __getitem__(self, item: int) -> dict[str, np.ndarray]:
        """Get a batch of pre-simulated data"""
        if not 0 <= item < self.num_batches:
            raise IndexError(f"Index {item} is out of bounds for dataset with {self.num_batches} batches.")

        item = slice(item * self.batch_size, (item + 1) * self.batch_size)
        item = self.indices[item]

        batch = {
            key: np.take(value, item, axis=0) if isinstance(value, np.ndarray) else value
            for key, value in self.data.items()
        }

        if self.adapter is not None:
            batch = self.adapter(batch, stage=self.stage)

        return batch

    @property
    def num_batches(self) -> int | None:
        return int(np.ceil(self.num_samples / self.batch_size))

    def on_epoch_end(self) -> None:
        self.shuffle()

    def shuffle(self) -> None:
        """Shuffle the dataset in-place."""
        np.random.shuffle(self.indices)

    @staticmethod
    def _get_num_samples_from_data(data: Mapping) -> int:
        for key, value in data.items():
            if hasattr(value, "shape"):
                ndim = len(value.shape)
                if ndim > 1:
                    return value.shape[0]

        raise ValueError("Could not determine number of samples from data. Please pass it manually.")



import pytest

from bayesflow.utils.serialization import deserialize, serializable, serialize


@serializable
class Foo:
    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        return {}


@serializable
class Bar:
    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    def get_config(self):
        return {}


@pytest.mark.parametrize("obj", [1, "1", True, False, None, ...])
def test_primitives(obj):
    assert deserialize(serialize(obj)) == obj


@pytest.mark.parametrize(
    "obj",
    [
        [1, 2, 3, "a", "b", None, True, ...],
        {"a": 1, "b": 2},
    ],
)
def test_collections_of_primitives(obj):
    assert deserialize(serialize(obj)) == obj


@pytest.mark.parametrize(
    "obj",
    [
        int,
        list,
    ],
)
def test_builtin_types(obj):
    assert deserialize(serialize(obj)) is obj


def test_custom_object():
    instance = Foo()

    assert isinstance(deserialize(serialize(instance)), Foo)


def test_custom_type():
    assert deserialize(serialize(Foo)) is Foo


@pytest.mark.parametrize(
    "obj",
    [
        [int, list, tuple, bool],
        {"a": int, "b": list, "c": tuple, "d": bool},
    ],
)
def test_collections_of_builtin_types(obj):
    assert deserialize(serialize(obj)) == obj


@pytest.mark.parametrize(
    "obj",
    [
        [Foo, Bar],
        {"a": Foo, "b": Bar},
    ],
)
def test_collection_of_custom_types(obj):
    assert deserialize(serialize(obj)) == obj



import keras

from bayesflow.types import Tensor
from bayesflow.utils import filter_kwargs, find_network, model_kwargs
from bayesflow.utils.serialization import deserialize, serializable, serialize

from ..invertible_layer import InvertibleLayer
from ..transforms import find_transform


@serializable
class SingleCoupling(InvertibleLayer):
    """
    Implements a single coupling layer as a composition of a subnet and a transform.

    Subnet output tensors are linearly mapped to the correct dimension.
    """

    MLP_DEFAULT_CONFIG = {
        "widths": (128, 128),
        "activation": "hard_silu",
        "kernel_initializer": "glorot_uniform",
        "residual": False,
        "dropout": 0.05,
        "spectral_normalization": False,
    }

    def __init__(
        self,
        subnet: str | type = "mlp",
        transform: str = "affine",
        subnet_kwargs: dict[str, any] = None,
        transform_kwargs: dict[str, any] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        subnet_kwargs = subnet_kwargs or {}
        transform_kwargs = transform_kwargs or {}

        if subnet == "mlp":
            subnet_kwargs = SingleCoupling.MLP_DEFAULT_CONFIG | subnet_kwargs

        self.subnet = find_network(subnet, **subnet_kwargs)
        self.transform = find_transform(transform, **transform_kwargs)

        self.output_projector = keras.layers.Dense(
            units=None, kernel_initializer="zeros", bias_initializer="zeros", name="output_projector"
        )

    def get_config(self):
        base_config = super().get_config()
        base_config = model_kwargs(base_config)

        config = {
            "subnet": self.subnet,
            "transform": self.transform,
            "output_projector": self.output_projector,
        }

        return base_config | serialize(config)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return cls(**deserialize(config, custom_objects=custom_objects))

    # noinspection PyMethodOverriding
    def build(self, x1_shape, x2_shape, conditions_shape=None):
        self.output_projector.units = self.transform.params_per_dim * x2_shape[-1]

        x1 = keras.ops.zeros(x1_shape)
        x2 = keras.ops.zeros(x2_shape)
        if conditions_shape is None:
            conditions = None
        else:
            conditions = keras.ops.zeros(conditions_shape)

        # build nested layers with forward pass
        self.call(x1, x2, conditions=conditions)

    def call(
        self, x1: Tensor, x2: Tensor, conditions: Tensor = None, inverse: bool = False, training: bool = False, **kwargs
    ) -> ((Tensor, Tensor), Tensor):
        if inverse:
            return self._inverse(x1, x2, conditions=conditions, training=training, **kwargs)
        return self._forward(x1, x2, conditions=conditions, training=training, **kwargs)

    def _forward(
        self, x1: Tensor, x2: Tensor, conditions: Tensor = None, training: bool = False, **kwargs
    ) -> ((Tensor, Tensor), Tensor):
        """Transform (x1, x2) -> (x1, f(x2; x1))"""
        z1 = x1
        parameters = self.get_parameters(x1, conditions=conditions, training=training)
        z2, log_det = self.transform(x2, parameters=parameters)

        return (z1, z2), log_det

    def _inverse(
        self, z1: Tensor, z2: Tensor, conditions: Tensor = None, training: bool = False, **kwargs
    ) -> ((Tensor, Tensor), Tensor):
        """Transform (x1, f(x2; x1)) -> (x1, x2)"""
        x1 = z1
        parameters = self.get_parameters(x1, conditions=conditions, training=training, **kwargs)
        x2, log_det = self.transform(z2, parameters=parameters, inverse=True)

        return (x1, x2), log_det

    def get_parameters(
        self, x: Tensor, conditions: Tensor = None, training: bool = False, **kwargs
    ) -> dict[str, Tensor]:
        """Applies the inner neural network to obtain the transformation parameters, for instance,
        if affine transformations, then [s, t] = NN(inputs), followed by a constraint, e.g., s = exp(s).
        """
        if conditions is not None:
            x = keras.ops.concatenate([x, conditions], axis=-1)

        parameters = self.output_projector(self.subnet(x, training=training, **filter_kwargs(kwargs, self.subnet.call)))
        parameters = self.transform.split_parameters(parameters)
        parameters = self.transform.constrain_parameters(parameters)

        return parameters



from collections.abc import Callable, Sequence, Mapping

import matplotlib.pyplot as plt
import numpy as np


from bayesflow.utils import prepare_plot_data, prettify_subplots, make_quadratic, add_titles_and_labels, add_metric


def recovery_from_estimates(
    estimates: Mapping[str, Mapping[str, np.ndarray]],
    targets: Mapping[str, np.ndarray],
    marker_mapping: dict[str, str],
    variable_keys: Sequence[str] = None,
    variable_names: Sequence[str] = None,
    add_corr: bool = True,
    corr_point_agg: Callable = np.median,
    figsize: Sequence[int] = None,
    label_fontsize: int = 16,
    title_fontsize: int = 18,
    metric_fontsize: int = 16,
    tick_fontsize: int = 12,
    color: str = "#132a70",
    num_col: int = None,
    num_row: int = None,
    xlabel: str = "Ground truth",
    ylabel: str = "Estimate",
    **kwargs,
) -> plt.Figure:
    """
    Creates and plots publication-ready recovery plot of estimates vs. targets.

    This plot yields similar information as the "posterior z-score",
    but allows for generic point and uncertainty estimates:

    https://betanalpha.github.io/assets/case_studies/principled_bayesian_workflow.html

    Important:
    Posterior aggregates play no special role in Bayesian inference and should only be used heuristically.
    For instance, in the case of multi-modal posteriors, common point estimates, such as mean, (geometric) median,
    or maximum a posteriori (MAP) need to be interpreted carefully.

    Parameters
    ----------
    estimates         : dict[str, dict[str, np.ndarray]]
        The model-generated estimates in a nested dictionary, e.g. as returned
        by approximator.estimate(conditions=...).
        - The outer keys identify the inference variable.
        - The inner keys identify point estimates.
        - The inner value is an ndarray of shape (num_datasets, point_estimate_size, variable_block_size)
    targets           : dict[str, np.ndarray]
        The prior draws (true parameters) used for generating the num_datasets
    marker_mapping    : dict[str, str]
        Define how to mark different point estimates by their key, e.g. {"quantiles":"_", "mean":"*"}.
        Only point estimates whose key appears in the dictionary, will be plotted.
    variable_keys     : list or None, optional, default: None
       Select keys from the dictionaries provided in estimates and targets.
       By default, select all keys.
    variable_names    : list or None, optional, default: None
        The individual parameter names for nice plot titles. Inferred if None
    add_corr          : boolean, default: True
        Should correlations between estimates and ground truth values be shown?
    corr_point_agg    : Callable
        Function producing a central point estimate from the whole list of point estimates
        in case correlations should be computed. Default: median
    figsize           : tuple or None, optional, default : None
        The figure size passed to the matplotlib constructor. Inferred if None.
    label_fontsize    : int, optional, default: 16
        The font size of the y-label text.
    title_fontsize    : int, optional, default: 18
        The font size of the title text.
    metric_fontsize   : int, optional, default: 16
        The font size of the metrics shown as text.
    tick_fontsize     : int, optional, default: 12
        The font size of the axis ticklabels.
    color             : str, optional, default: '#8f2727'
        The color for the true vs. estimated scatter points and error bars.
    num_row           : int, optional, default: None
        The number of rows for the subplots. Dynamically determined if None.
    num_col           : int, optional, default: None
        The number of columns for the subplots. Dynamically determined if None.
    xlabel:
    ylabel:

    Returns
    -------
    f : plt.Figure - the figure instance for optional saving

    Raises
    ------
    ShapeError
        If there is a deviation from the expected shapes of ``estimates`` and ``targets``.
    """

    point_estimates = {}
    markers = []
    for var_key, var_value in estimates.items():
        estimate_list = []
        for estimate_key in marker_mapping.keys():
            e = estimates[var_key][estimate_key]
            # move axis of flexible shape to the end
            e = np.moveaxis(e, 1, -1)
            # add dimensions of size (1,) if not 3D yet
            e = np.atleast_3d(e)
            # undo reordering
            e = np.moveaxis(e, -1, 1)
            estimate_list.append(e)
            markers += [marker_mapping[estimate_key]] * e.shape[1]
        point_estimates[var_key] = np.concatenate(estimate_list, axis=1)

    # Gather plot data and metadata into a dictionary
    plot_data = prepare_plot_data(
        estimates=point_estimates,
        targets=targets,
        variable_keys=variable_keys,
        variable_names=variable_names,
        num_col=num_col,
        num_row=num_row,
        figsize=figsize,
    )

    estimates = plot_data.pop("estimates")
    targets = plot_data.pop("targets")

    # Compute point estimates and uncertainties
    point_estimate = corr_point_agg(estimates, axis=1)

    num_quantiles = estimates.shape[1]

    for i, ax in enumerate(plot_data["axes"].flat):
        if i >= plot_data["num_variables"]:
            break

        # Add scatter and error bars
        for q_idx in range(num_quantiles):
            _ = ax.scatter(
                targets[:, i],
                estimates[:, q_idx, i],
                marker=markers[q_idx],
                alpha=0.5,
                color=color,
                **kwargs,
            )

        connecting_lines_x = np.tile(targets[:, i], (2, 1))
        connecting_lines_y = np.array([estimates[:, :, i].min(axis=1), estimates[:, :, i].max(axis=1)])
        _ = ax.plot(
            connecting_lines_x,
            connecting_lines_y,
            alpha=0.5,
            color=color,
        )

        make_quadratic(ax, targets[:, i], point_estimate[:, i])

        if add_corr:
            corr = np.corrcoef(targets[:, i], point_estimate[:, i])[0, 1]
            add_metric(ax=ax, metric_text="$r$", metric_value=corr, metric_fontsize=metric_fontsize)

        ax.set_title(plot_data["variable_names"][i], fontsize=title_fontsize)

    # Add custom schmuck
    prettify_subplots(plot_data["axes"], num_subplots=plot_data["num_variables"], tick_fontsize=tick_fontsize)
    add_titles_and_labels(
        axes=plot_data["axes"],
        num_row=plot_data["num_row"],
        num_col=plot_data["num_col"],
        xlabel=xlabel,
        ylabel=ylabel,
        label_fontsize=label_fontsize,
    )

    plot_data["fig"].tight_layout()
    return plot_data["fig"]



import keras

from bayesflow.types import Tensor
from bayesflow.utils import check_lengths_same
from bayesflow.utils.serialization import serializable

from ..embeddings import Time2Vec, RecurrentEmbedding
from ..summary_network import SummaryNetwork

from .mab import MultiHeadAttentionBlock


@serializable
class TimeSeriesTransformer(SummaryNetwork):
    def __init__(
        self,
        summary_dim: int = 16,
        embed_dims: tuple = (64, 64),
        num_heads: tuple = (4, 4),
        mlp_depths: tuple = (2, 2),
        mlp_widths: tuple = (128, 128),
        dropout: float = 0.05,
        mlp_activation: str = "gelu",
        kernel_initializer: str = "lecun_normal",
        use_bias: bool = True,
        layer_norm: bool = True,
        time_embedding: str = "time2vec",
        time_embed_dim: int = 8,
        time_axis: int = None,
        **kwargs,
    ):
        """Creates a regular transformer coupled with Time2Vec embeddings of time used to flexibly compress time series.
        If the time intervals vary across batches, it is highly recommended that your simulator also returns a "time"
        vector appended to the simulator outputs and specified via the "time_axis" argument.

        Parameters
        ----------
        summary_dim : int, optional (default - 16)
            Dimensionality of the final summary output.
        embed_dims  : tuple of int, optional (default - (64, 64))
            Dimensions of the keys, values, and queries for each attention block.
        num_heads   : tuple of int, optional (default - (4, 4))
            Number of attention heads for each embedding dimension.
        mlp_depths  : tuple of int, optional (default - (2, 2))
            Depth of the multi-layer perceptron (MLP) blocks for each component.
        mlp_widths  : tuple of int, optional (default - (128, 128))
            Width of each MLP layer in each block for each component.
        dropout     : float, optional (default - 0.05)
            Dropout rate applied to the attention and MLP layers. If set to None, no dropout is applied.
        mlp_activation : str, optional (default - 'gelu')
            Activation function used in the dense layers. Common choices include "relu", "elu", and "gelu".
        kernel_initializer : str, optional (default - 'lecun_normal')
            Initializer for the kernel weights matrix. Common choices include "glorot_uniform", "he_normal", etc.
        use_bias : bool, optional (default - True)
            Whether to include a bias term in the dense layers.
        layer_norm : bool, optional (default - True)
            Whether to apply layer normalization after the attention and MLP layers.
        time_embedding  : str, optional (default - "time2vec")
            The type of embedding to use. Must be in ["time2vec", "lstm", "gru"]
        time_embed_dim  : int, optional (default - 8)
            The dimensionality of the Time2Vec or recurrent embedding.
        time_axis     : int, optional (default - None)
            The time axis (e.g., -1 for last axis) from which to grab the time vector that goes into the embedding.
            If an embedding is provided and time_axis is None, a uniform time interval between [0, sequence_len]
            will be assumed.
        **kwargs : dict
            Additional keyword arguments passed to the base layer.
        """

        super().__init__(**kwargs)

        # Ensure all tuple-settings have the same length
        check_lengths_same(embed_dims, num_heads, mlp_depths, mlp_widths)

        # Initialize Time2Vec embedding layer
        if time_embedding is None:
            self.time_embedding = None
        elif time_embedding == "time2vec":
            self.time_embedding = Time2Vec(num_periodic_features=time_embed_dim - 1)
        elif time_embedding in ["lstm", "gru"]:
            self.time_embedding = RecurrentEmbedding(time_embed_dim, time_embedding)
        else:
            raise ValueError("Embedding not found!")

        # Construct a series of set-attention blocks
        self.attention_blocks = []
        for i in range(len(embed_dims)):
            layer_attention_settings = dict(
                dropout=dropout,
                mlp_activation=mlp_activation,
                kernel_initializer=kernel_initializer,
                use_bias=use_bias,
                layer_norm=layer_norm,
                num_heads=num_heads[i],
                embed_dim=embed_dims[i],
                mlp_depth=mlp_depths[i],
                mlp_width=mlp_widths[i],
            )

            block = MultiHeadAttentionBlock(**layer_attention_settings)
            self.attention_blocks.append(block)

        # Pooling will be applied as a final step to the abstract representations obtained from set attention
        self.pooling = keras.layers.GlobalAvgPool1D()
        self.output_projector = keras.layers.Dense(summary_dim)

        self.summary_dim = summary_dim
        self.time_axis = time_axis

    def call(self, input_sequence: Tensor, training: bool = False, **kwargs) -> Tensor:
        """Compresses the input sequence into a summary vector of size `summary_dim`.

        Parameters
        ----------
        input_sequence  : Tensor
            Input of shape (batch_size, sequence_length, input_dim)
        training        : boolean, optional (default - False)
            Passed to the optional internal dropout and spectral normalization
            layers to distinguish between train and test time behavior.
        **kwargs        : dict, optional (default - {})
            Additional keyword arguments passed to the internal attention layer,
            such as ``attention_mask`` or ``return_attention_scores``

        Returns
        -------
        out : Tensor
            Output of shape (batch_size, set_size, output_dim)
        """

        if self.time_axis is not None:
            t = input_sequence[..., self.time_axis]
            indices = list(range(keras.ops.shape(input_sequence)[-1]))
            indices.pop(self.time_axis)
            inp = keras.ops.take(input_sequence, indices, axis=-1)
        else:
            t = None
            inp = input_sequence

        if self.time_embedding:
            inp = self.time_embedding(inp, t=t)

        # Apply self-attention blocks
        for layer in self.attention_blocks:
            inp = layer(inp, inp, training=training, **kwargs)

        # Global average pooling and output projection
        summary = self.pooling(inp)
        summary = self.output_projector(summary)
        return summary
